/*
 * Copyright 2016 ELVEES NeoTek JSC
 * Copyright 2017 RnD Center "ELVEES", JSC
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <asm/uaccess.h>
#include <linux/anon_inodes.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/dma-buf.h>
#include <linux/fs.h>
#include <linux/genalloc.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/kthread.h>
#include <linux/module.h>
#include <linux/of.h>
#include <linux/of_address.h>
#include <linux/of_platform.h>
#include <linux/poll.h>
#include <linux/regmap.h>
#include <linux/reset.h>
#include <linux/sched.h>
#include <linux/sizes.h>
#include <linux/slab.h>
#include <linux/mfd/syscon.h>

#include <linux/vdma.h>

#include "elcore40.h"
#include "regs.h"

struct region {
	struct resource res;
	struct mutex requested;
};

struct elcore_private_data {
	void __iomem *regs;

	struct device *dev;

	struct gen_pool *vpool;
	struct region codebuf;
	struct list_head job_queue;
	spinlock_t lock;

	/* bitmaps for Elcore40 resources */
	DECLARE_BITMAP(mcores, MAX_CORES);
	DECLARE_BITMAP(mvdma, MAX_VDMAS);
	DECLARE_BITMAP(mevents, MAX_EVENTS);

	dev_t dev_num;
	struct class *class;
	struct cdev cdev;
};

/**
 * struct elcore_buffer_desc - Internal data about allocated buffer
 * @pdata: Pointer to driver private data.
 * @buf:   Embedded user-visible buffer data.
 * @paddr: Physical address of allocated memory.
 * @vaddr: Virtual address of allocated memory in kernel address space.
 */
struct elcore_buffer_desc {
	struct elcore_private_data *pdata;

	struct elcore_buffer buf;
	struct dma_buf *dma_buf;

	dma_addr_t paddr;
	void *vaddr;
};

/**
 * struct elcore_program_args - Arguments to Elcore40 code
 * @inum: Number of input buffers.
 * @onum: Number of output buffers.
 * @bufs: Physical addresses of input buffers followed by output buffer.
 */
struct elcore_program_args {
	int inum;
	int onum;
	phys_addr_t bufs[];
};

/**
 * struct buf_info - Information on imported buffer
 * @dmabuf: Pointer to DMA buffer description associated with buffer.
 * @attach: DMA buffer attachment.
 * @sgt:    SG table returned after attachment mapping.
 */
struct buf_info {
	struct dma_buf_attachment *attach;
	struct sg_table *sgt;
};

/**
 * struct elcore_job_desc - Internal data about created job
 * @pdata: Pointer to driver private data.
 * @job:   Embedded user-visible job data.
 * @args:  Pointer to &struct elcore_program_args to be passed to job code.
 * @kernel: Array of &struct buf_info for kernel buffers.
 * @input:  Array of &struct buf_info for input buffers.
 * @output: Array of &struct buf_info for output buffers.
 * @stack:  Physical and virtual (kernel address space) addresses of job stack.
 * @job_queue_node: List of jobs enqueued to be run.
 * @lock:   Lock for job status.
 * @program_entry: Entry point for program executed on Elcore40.
 */
struct elcore_job_desc {
	struct elcore_private_data *pdata;
	struct elcore_job job;
	struct elcore_buffer_desc args;

	struct buf_info kernel[MAX_KERNELS];
	struct buf_info input[MAX_INPUTS];
	struct buf_info output[MAX_OUTPUTS];

	struct elcore_stack {
		dma_addr_t paddr;
		void *vaddr;
	} stack;

	struct list_head job_queue_node;
	spinlock_t lock;

	u32 program_entry;
};

/* FIXME: These dividers shouldn't be set here. This is the temporary solution
 * to ease frequency change. */
static unsigned int sys3div = 8;
module_param(sys3div, uint, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(sys3div, "SYS3 divider (default: 8)");

static unsigned int cip3div = 4;
module_param(cip3div, uint, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(sys3div, "CIP3 divider (default: 4)");

static int elcore_mem_alloc(struct elcore_buffer_desc *buf_desc)
{
	/* The size of allocated buffer should be aligned to page size,
	 * because user application may want to mmap it.
	 */
	buf_desc->buf.size = ALIGN(buf_desc->buf.size, PAGE_SIZE);

	switch (buf_desc->buf.type) {
	case ELCORE_MEMORY_CODE:
		if (buf_desc->buf.size >
		    resource_size(&buf_desc->pdata->codebuf.res))
			return -ENOMEM;

		if (!mutex_trylock(&buf_desc->pdata->codebuf.requested))
			return -EBUSY;

		buf_desc->paddr = buf_desc->pdata->codebuf.res.start;
		buf_desc->vaddr = NULL;

		break;
	case ELCORE_MEMORY_VRAM:
		buf_desc->vaddr = gen_pool_dma_alloc(buf_desc->pdata->vpool,
						     buf_desc->buf.size,
						     &buf_desc->paddr);
		if (!buf_desc->vaddr) {
			dev_err(buf_desc->pdata->dev,
				"VRAM allocation failed\n");
			return -ENOMEM;
		}

		dev_dbg(buf_desc->pdata->dev,
			"%u@0x%08lx bytes of VRAM allocated\n",
			buf_desc->buf.size, (unsigned long)buf_desc->vaddr);

		break;
	case ELCORE_MEMORY_SYSTEM:
		buf_desc->vaddr = dma_alloc_coherent(buf_desc->pdata->dev,
						     buf_desc->buf.size,
						     &buf_desc->paddr,
						     GFP_KERNEL);
		if (!buf_desc->vaddr) {
			dev_err(buf_desc->pdata->dev,
				"System memory allocation failed\n");
			return -ENOMEM;
		}

		dev_dbg(buf_desc->pdata->dev,
			"%u@0x%08lx bytes of SYSTEM MEMORY allocated\n",
			buf_desc->buf.size, (unsigned long)buf_desc->vaddr);

		break;
	default:
		dev_err(buf_desc->pdata->dev, "Unknown memory type\n");
		return -EINVAL;
	}

	return 0;
}

static void elcore_mem_free(struct elcore_buffer_desc *buf_desc)
{
	switch (buf_desc->buf.type) {
	case ELCORE_MEMORY_CODE:
		mutex_unlock(&buf_desc->pdata->codebuf.requested);
		break;
	case ELCORE_MEMORY_VRAM:
		gen_pool_free(buf_desc->pdata->vpool,
			      (unsigned long)buf_desc->vaddr,
			      buf_desc->buf.size);

		dev_dbg(buf_desc->pdata->dev,
			"%u@0x%08lx bytes of VRAM freed\n",
			buf_desc->buf.size,
			(unsigned long)buf_desc->vaddr);

		break;
	case ELCORE_MEMORY_SYSTEM:
		dma_free_coherent(buf_desc->pdata->dev, buf_desc->buf.size,
				  buf_desc->vaddr, buf_desc->paddr);

		dev_dbg(buf_desc->pdata->dev,
			"%u@0x%08lx bytes of SYSTEM MEMORY freed\n",
			buf_desc->buf.size, (unsigned long)buf_desc->vaddr);

		break;
	default:
		dev_warn(buf_desc->pdata->dev, "Unsupported buffer type %d\n",
			 buf_desc->buf.type);
	}
}

static struct sg_table *elcore_dmabuf_map(struct dma_buf_attachment *attach,
					  enum dma_data_direction dir)
{
	int rc;
	struct sg_table *sgt;

	/* I'm sure attach is not NULL and corresponds to my dma_buf. */
	struct elcore_buffer_desc *desc = attach->dmabuf->priv;

	/* TODO: dmabuf-specific lock */

	sgt = kmalloc(sizeof(*sgt), GFP_KERNEL);
	if (!sgt)
		return ERR_PTR(-ENOMEM);

	rc = sg_alloc_table(sgt, 1, GFP_KERNEL);
	if (rc)
		goto free_sgt;

	/* FIXME: This will not work for CODE buffer, because its vaddr
	 * is NULL. */
	sg_init_one(sgt->sgl, desc->vaddr, desc->buf.size);

	rc = dma_map_sg(attach->dev, sgt->sgl, sgt->orig_nents, dir);
	if (rc <= 0) {
		pr_err("Failed to map DMA buffer\n");
		goto free_table;
	}

	/* FIXME: dma_map_sg() doesn't determine physical addresses of buffers
	 * correctly in all cases. For example it doesn't work for gen_pool
	 * buffers. One possibility to deal with it correctly would be to fix
	 * dma_map_sg(). Another (possibly better) is to create device-specific
	 * dma_ops. */

	sgt->sgl[0].dma_address = desc->paddr;

	return sgt;

free_table:
	sg_free_table(sgt);
free_sgt:
	kfree(sgt);
	return ERR_PTR(rc);
}

static void elcore_dmabuf_unmap(struct dma_buf_attachment *attach,
				struct sg_table *sgt,
				enum dma_data_direction dir)
{
	dma_unmap_sg(attach->dev, sgt->sgl, sgt->orig_nents, dir);
	sg_free_table(sgt);
	kfree(sgt);
}

static void *elcore_dmabuf_kmap_atomic(struct dma_buf *dma_buf,
					unsigned long page_num)
{
	/* TODO */
	return NULL;
}

static void *elcore_dmabuf_kmap(struct dma_buf *dma_buf,
				 unsigned long page_num)
{
	/* TODO */
	return NULL;
}

/*
 * FIXME: This function doesn't work correctly for buffer whose address is not
 * aligned to the page boundary. It should either return an error or handle
 * this in some standard way.
 */
static int elcore_dmabuf_mmap(struct dma_buf *dma_buf,
			       struct vm_area_struct *vma)
{
	struct elcore_buffer_desc *buf_desc = dma_buf->priv;
	struct elcore_private_data *pdata = buf_desc->pdata;
	int ret;

	if (vma->vm_pgoff) {
		dev_err(pdata->dev, "Nonzero vm_pgoff is not supported\n");
		return -EINVAL;
	}

	switch (buf_desc->buf.type) {
	case ELCORE_MEMORY_CODE:
	case ELCORE_MEMORY_VRAM:
		vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
		/*
		 * TODO: Should this be io_remap_pfn_range() or
		 * vm_iomap_memory()?
		 */
		ret = remap_pfn_range(vma, vma->vm_start,
				      buf_desc->paddr >> PAGE_SHIFT,
				      vma->vm_end - vma->vm_start,
				      vma->vm_page_prot);
		if (ret)
			goto err1;
		break;
	case ELCORE_MEMORY_SYSTEM:
		ret = dma_mmap_coherent(pdata->dev, vma,
					buf_desc->vaddr, buf_desc->paddr,
					buf_desc->buf.size);

		if (ret)
			goto err1;
		break;
	}

	dev_dbg(pdata->dev, "0x%08lx dma address mapped to 0x%08lx\n",
		(unsigned long)buf_desc->paddr, vma->vm_start);

	return 0;

err1:
	dev_err(pdata->dev, "Memory mapping failed\n");
	return ret;
}

static void elcore_dmabuf_release(struct dma_buf *dmabuf)
{
	struct elcore_buffer_desc *buf_desc = dmabuf->priv;

	elcore_mem_free(buf_desc);
	devm_kfree(buf_desc->pdata->dev, buf_desc);
}

static struct dma_buf_ops elcore_dmabuf_ops = {
	.map_dma_buf = elcore_dmabuf_map,
	.unmap_dma_buf = elcore_dmabuf_unmap,
	.kmap = elcore_dmabuf_kmap,
	.kmap_atomic = elcore_dmabuf_kmap_atomic,
	.mmap = elcore_dmabuf_mmap,
	.release = elcore_dmabuf_release
};

/*
 * BUG: This is clearly a dirty-dirty hack, because this will not work
 * for an imported dma_buf. Importer should not use &dma_buf.priv, it is
 * exporter private data.
 */
static struct elcore_buffer_desc *elcore_buf_desc_get(struct device *const dev,
						      int const fd)
{
	struct dma_buf *dma_buf;

	dma_buf = dma_buf_get(fd);
	if (IS_ERR(dma_buf)) {
		dev_err(dev, "Can not get dma_buf for fd %d\n", fd);
		return (void *)dma_buf;
	}

	if (dma_buf->ops != &elcore_dmabuf_ops) {
		dev_err(dev, "Work with imported buffer is not supported");
		dma_buf_put(dma_buf);
		return ERR_PTR(-EINVAL);
	}

	return dma_buf->priv;
}

static void elcore_buf_desc_put(struct elcore_buffer_desc *const buf_desc)
{
	if (buf_desc)
		dma_buf_put(buf_desc->dma_buf);
}

static int elcore_resource_alloc(enum elcore_resource_type type,
				 struct elcore_private_data *pdata)
{
	int idx = 0;
	char *res_name;

	switch (type) {
	case ELCORE_CORE:
		res_name = "CORE";
		idx = find_first_zero_bit(pdata->mcores, MAX_CORES);
		if (idx == MAX_CORES)
			goto err1;

		set_bit(idx, pdata->mcores);
		break;
	case ELCORE_VDMA:
		res_name = "VDMA";
		idx = find_first_zero_bit(pdata->mvdma, MAX_VDMAS);
		if (idx == MAX_VDMAS)
			goto err1;

		set_bit(idx, pdata->mvdma);
		break;
	case ELCORE_EVENT:
		res_name = "EVENT";
		idx = find_first_zero_bit(pdata->mevents, MAX_EVENTS);
		if (idx == MAX_EVENTS)
			goto err1;

		set_bit(idx, pdata->mevents);
		break;
	default:
		dev_err(pdata->dev, "Unknown resource type\n");
		return 0;
	}

	dev_dbg(pdata->dev, "%s%d resource claimed", res_name, idx);
	return idx;

err1:
	dev_err(pdata->dev, "Can't find free %s resource\n", res_name);
	return -1;
}

static void elcore_resource_free(struct elcore_resource *res,
				 struct elcore_private_data *pdata)
{
	char *res_name;

	switch (res->type) {
	case ELCORE_CORE:
		res_name = "CORE";
		clear_bit(res->descriptor, pdata->mcores);
		break;
	case ELCORE_VDMA:
		res_name = "VDMA";
		clear_bit(res->descriptor, pdata->mvdma);
		break;
	case ELCORE_EVENT:
		res_name = "EVENT";
		clear_bit(res->descriptor, pdata->mevents);
		break;
	default:
		dev_err(pdata->dev, "Unknown resource type\n");
		return;
	}

	dev_dbg(pdata->dev, "%s%d resource freed", res_name, res->descriptor);
}

static int buf_info_attach(struct device *dev, struct buf_info *info, int fd)
{
	struct dma_buf *dmabuf = dma_buf_get(fd);

	if (IS_ERR(dmabuf)) {
		dev_err(dev, "Failed to get DMA buffer for fd %d\n", fd);
		return PTR_ERR(dmabuf);
	}

	info->attach = dma_buf_attach(dmabuf, dev);
	if (IS_ERR(info->attach)) {
		dev_err(dev, "Failed to attach DMA buffer for fd %d\n", fd);
		dma_buf_put(dmabuf);
		return PTR_ERR(info->attach);
	}

	return 0;
}

static void buf_info_detach(struct device *dev, struct buf_info *info)
{
	struct dma_buf *dmabuf;

	if (!IS_ERR_OR_NULL(info->attach)) {
		dmabuf = info->attach->dmabuf;

		dma_buf_detach(dmabuf, info->attach);
		info->attach = NULL;

		dma_buf_put(dmabuf);
	}
}

static void detach_buffers(struct elcore_job_desc *desc)
{
	int i;
	struct device *dev = desc->pdata->dev;

	for (i = 0; i < desc->job.onum; ++i)
		buf_info_detach(dev, &desc->output[i]);

	for (i = 0; i < desc->job.inum; ++i)
		buf_info_detach(dev, &desc->input[i]);

	buf_info_detach(dev, &desc->kernel[0]);
}

static int attach_buffers(struct elcore_job_desc *desc)
{
	int rc, i;
	struct device *dev = desc->pdata->dev;

	rc = buf_info_attach(dev, &desc->kernel[0], desc->job.kernel[0]);
	if (rc)
		goto err;

	for (i = 0; i < desc->job.inum; ++i) {
		rc = buf_info_attach(dev, &desc->input[i], desc->job.input[i]);
		if (rc)
			goto err;
	}

	for (i = 0; i < desc->job.onum; ++i) {
		rc = buf_info_attach(dev, &desc->output[i],
				     desc->job.output[i]);
		if (rc)
			goto err;
	}

	return 0;

err:
	detach_buffers(desc);
	return rc;
}

static int buf_info_map(struct buf_info *info, enum dma_data_direction dir)
{
	/* TODO: Sanity check. To be removed. */
	if (WARN_ON(IS_ERR_OR_NULL(info->attach)))
		return -EINVAL;

	info->sgt = dma_buf_map_attachment(info->attach, dir);
	if (IS_ERR(info->sgt))
		/* TODO: Error message? */
		return PTR_ERR(info->sgt);

	return 0;
}

static void buf_info_unmap(struct buf_info *info, enum dma_data_direction dir)
{
	if (!IS_ERR_OR_NULL(info->sgt)) {
		dma_buf_unmap_attachment(info->attach, info->sgt, dir);
		info->sgt = NULL;
	}
}

static void unmap_buffers(struct elcore_job_desc *desc)
{
	int i;

	for (i = 0; i < desc->job.onum; ++i)
		buf_info_unmap(&desc->output[i], DMA_BIDIRECTIONAL);

	for (i = 0; i < desc->job.inum; ++i)
		buf_info_unmap(&desc->input[i], DMA_TO_DEVICE);

	buf_info_unmap(&desc->kernel[0], DMA_TO_DEVICE);
}

static int map_buffers(struct elcore_job_desc *desc)
{
	int rc, i;

	rc = buf_info_map(&desc->kernel[0], DMA_TO_DEVICE);
	if (rc)
		goto err;

	for (i = 0; i < desc->job.inum; ++i) {
		rc = buf_info_map(&desc->input[i], DMA_TO_DEVICE);
		if (rc)
			goto err;
	}

	for (i = 0; i < desc->job.onum; ++i) {
		/* DMA_BIDIRECTIONAL is used, because libvi sometimes use output
		 * buffers as bidirectional. */
		/* TODO: Rework to use single buffer list in elcore_job */
		rc = buf_info_map(&desc->output[i], DMA_BIDIRECTIONAL);
		if (rc)
			goto err;
	}

	return 0;

err:
	unmap_buffers(desc);
	return rc;
}

static int elcore_job_release(struct inode *inode, struct file *file)
{
	struct elcore_job_desc *desc = file->private_data;

	if (desc->job.status != ELCORE_JOB_IDLE)
		return -EBUSY;

	if (desc->job.flags & ELCORE_INIT_STACK)
		gen_pool_free(desc->pdata->vpool,
			      (unsigned long)desc->stack.vaddr, PAGE_SIZE);

	unmap_buffers(desc);
	detach_buffers(desc);

	elcore_mem_free(&desc->args);

	kfree(desc);

	return 0;
}

static unsigned int elcore_job_poll(struct file *file, poll_table *wait)
{
	return 0;
}

static const struct file_operations elcore_job_fops = {
	.release = elcore_job_release,
	.poll = elcore_job_poll
};

static void elcore_run_job(struct elcore_job_desc *job_desc)
{
	struct elcore_private_data *pdata = job_desc->pdata;
	unsigned long flags;

	WARN_ON(elcore_readl(pdata, CORE0_DCSR) & CORE0_DCSR_RUN);

	/* TODO: I think it is bad idea to printk() under spinlock */
	dev_dbg(pdata->dev, "elcore job @ %p is got\n", job_desc);

	spin_lock_irqsave(&job_desc->lock, flags);
	if (job_desc->job.status != ELCORE_JOB_ENQUEUED) {
		dev_err(pdata->dev, "wrong job status (%d)\n",
			job_desc->job.status);
		spin_unlock_irqrestore(&job_desc->lock, flags);
		return;
	}

	/* Mark job as running */
	job_desc->job.status = ELCORE_JOB_RUNNING;
	spin_unlock_irqrestore(&job_desc->lock, flags);

	elcore_writel(pdata, CORE0_R0, job_desc->args.paddr);
	elcore_writel(pdata, CORE0_PC, job_desc->program_entry);

	if (job_desc->job.flags & ELCORE_INIT_STACK)
		elcore_writel(pdata, CORE0_A7,
			      job_desc->stack.paddr + PAGE_SIZE);

	elcore_writel(pdata, CORE0_DCSR, CORE0_DCSR_RUN);
}

static irqreturn_t elcore_interrupt(int irq, void *arg)
{
	struct elcore_private_data *pdata = arg;
	struct elcore_job_desc *job_desc;
	unsigned long flags;
	u32 val;

	/* HWSGBM interrupts should only be handled by the HWSGBM driver */
	if (!(elcore_readl(pdata, QSTR_ELCORE) & ~HWSGBM_INTS))
		return IRQ_NONE;

	spin_lock_irqsave(&pdata->lock, flags);
	job_desc = list_first_entry(&pdata->job_queue,
				    struct elcore_job_desc,
				    job_queue_node);
	spin_unlock_irqrestore(&pdata->lock, flags);

	val = elcore_readl(pdata, CORE0_DCSR);
	if (!(val & CORE0_DCSR_PI) && !(val & CORE0_DCSR_SE) &&
	    !(val & CORE0_DCSR_BRK) && (val & CORE0_DCSR_STOP))
		job_desc->job.rc = ELCORE_JOB_SUCCESS;
	else
		job_desc->job.rc = ELCORE_JOB_ERROR;

	/* There is no need to protect job.status here since the
	 * following line is the only place where status can be changed
	 * from ELCORE_JOB_RUNNING state.
	 */
	job_desc->job.status = ELCORE_JOB_IDLE;

	elcore_writel(pdata, CORE0_DQSTR, 0);
	elcore_writel(pdata, QSTR_ELCORE, 0);
	elcore_writel(pdata, CORE0_DCSR, 0);

	spin_lock_irqsave(&pdata->lock, flags);
	list_del(&job_desc->job_queue_node);

	if (!list_empty(&pdata->job_queue)) {
		job_desc = list_first_entry(&pdata->job_queue,
					    struct elcore_job_desc,
					    job_queue_node);
		/* TODO: Need to garantee that job is in valid state */
		elcore_run_job(job_desc);
	}
	spin_unlock_irqrestore(&pdata->lock, flags);

	/* TODO: Notify poll() that job is completed */

	return IRQ_HANDLED;
}

static int elcore_buf_alloc(struct elcore_private_data *pdata,
			    void __user *arg)
{
	struct dma_buf *dmabuf;
	struct elcore_buffer_desc *buf_desc;
	DEFINE_DMA_BUF_EXPORT_INFO(exp_info);
	int ret;

	buf_desc = devm_kzalloc(pdata->dev, sizeof(struct elcore_buffer_desc),
				GFP_KERNEL);
	if (!buf_desc)
		return -ENOMEM;

	ret = copy_from_user(&buf_desc->buf, arg, sizeof(struct elcore_buffer));
	if (ret) {
		ret = -EACCES;
		goto err_free_buf_desc;
	}

	buf_desc->pdata = pdata;

	ret = elcore_mem_alloc(buf_desc);
	if (ret)
		goto err_free_buf_desc;

	exp_info.ops = &elcore_dmabuf_ops;
	exp_info.size = buf_desc->buf.size;
	exp_info.flags = O_RDWR;
	exp_info.priv = buf_desc;

	dmabuf = dma_buf_export(&exp_info);
	if (IS_ERR(dmabuf)) {
		ret = PTR_ERR(dmabuf);
		goto err_free_mem;
	}

	buf_desc->dma_buf = dmabuf;
	buf_desc->buf.fd = dma_buf_fd(dmabuf, 0);
	if (!buf_desc->buf.fd) {
		/* This will call elcore_dma_buf_release() */
		dma_buf_put(dmabuf);
		return -EINVAL;
	}

	ret = copy_to_user(arg, &buf_desc->buf, sizeof(struct elcore_buffer));
	if (ret) {
		dma_buf_put(dmabuf);
		return -EFAULT;
	}

	return 0;

err_free_mem:
	if (buf_desc->buf.type == ELCORE_MEMORY_VRAM)
		gen_pool_free(buf_desc->pdata->vpool,
			      (unsigned long)buf_desc->vaddr,
			      buf_desc->buf.size);

	if (buf_desc->buf.type == ELCORE_MEMORY_SYSTEM)
		dma_free_coherent(buf_desc->pdata->dev,
				  buf_desc->buf.size,
				  buf_desc->vaddr, buf_desc->paddr);
err_free_buf_desc:
	devm_kfree(pdata->dev, buf_desc);
	return ret;
}

static int elcore_resource_request(struct elcore_private_data *pdata,
				   void __user *arg)
{
	struct elcore_resource res;
	int ret;

	ret = copy_from_user(&res, arg, sizeof(struct elcore_resource));
	if (ret)
		return -EACCES;

	res.descriptor = elcore_resource_alloc(res.type, pdata);
	if (res.descriptor == -1)
		return -EBUSY;

	ret = copy_to_user(arg, &res, sizeof(struct elcore_resource));
	if (ret) {
		elcore_resource_free(&res, pdata);
		return -EFAULT;
	}

	return 0;
}

static int elcore_resource_release(struct elcore_private_data *pdata,
				   void __user *arg)
{
	struct elcore_resource res;
	int ret;

	ret = copy_from_user(&res, arg, sizeof(struct elcore_resource));
	if (ret)
		return -EACCES;

	elcore_resource_free(&res, pdata);

	return 0;
}

static int set_args(struct elcore_job_desc *desc)
{
	struct elcore_program_args *args;
	int rc, i;

	/* TODO: It seems redundant to have two pdata in one struct */
	desc->args.pdata = desc->pdata;
	desc->args.buf = (struct elcore_buffer){
		.type = ELCORE_MEMORY_VRAM,
		.size = sizeof(struct elcore_program_args) +
			desc->job.inum * sizeof(dma_addr_t) +
			desc->job.onum * sizeof(dma_addr_t)
	};

	rc = elcore_mem_alloc(&desc->args);
	if (rc)
		return rc;

	args = desc->args.vaddr;

	args->inum = desc->job.inum;
	args->onum = desc->job.onum;

	/* Assume that Elcore40 program entry point is located at the beginning
	 * of first kernel. */
	desc->program_entry = sg_dma_address(desc->kernel[0].sgt->sgl);

	for (i = 0; i < args->inum; ++i)
		args->bufs[i] = sg_dma_address(desc->input[i].sgt->sgl);

	for (i = 0; i < args->onum; ++i)
		args->bufs[i + args->inum] =
			sg_dma_address(desc->output[i].sgt->sgl);

	return 0;
}

static int elcore_job_create(struct elcore_private_data *pdata,
			     void __user *arg)
{
	struct elcore_job_desc *job_desc;
	struct device *dev = pdata->dev;
	int ret;

	job_desc = kzalloc(sizeof(struct elcore_job_desc), GFP_KERNEL);
	if (!job_desc)
		return -ENOMEM;

	job_desc->pdata = pdata;

	ret = copy_from_user(&job_desc->job, arg, sizeof(struct elcore_job));
	if (ret) {
		ret = -EACCES;
		goto err_desc;
	}

	spin_lock_init(&job_desc->lock);

	ret = anon_inode_getfd("elcorejob", &elcore_job_fops, job_desc, 0);
	if (ret < 0)
		goto err_desc;

	job_desc->job.fd = ret;
	job_desc->job.status = ELCORE_JOB_IDLE;

	ret = copy_to_user(arg, &job_desc->job, sizeof(struct elcore_job));
	if (ret) {
		ret = -EFAULT;
		goto err_fd;
	}

	if (job_desc->job.inum > MAX_INPUTS) {
		dev_err(dev, "Number of input buffers is too big (%d)\n",
			job_desc->job.inum);
		ret = -EINVAL;
		goto err_fd;
	}

	if (job_desc->job.onum > MAX_OUTPUTS) {
		dev_err(dev, "Number of output buffers is too big (%d)\n",
			job_desc->job.onum);
		ret = -EINVAL;
		goto err_fd;
	}

	ret = attach_buffers(job_desc);
	if (ret)
		goto err_fd;

	ret = map_buffers(job_desc);
	if (ret)
		goto detach_buffers;

	ret = set_args(job_desc);
	if (ret)
		goto unmap_buffers;

	if (job_desc->job.flags & ELCORE_INIT_STACK) {
		struct elcore_stack *stack = &job_desc->stack;

		stack->vaddr = gen_pool_dma_alloc(pdata->vpool, PAGE_SIZE,
						  &stack->paddr);
		if (!stack->vaddr) {
			dev_err(pdata->dev, "Failed to allocate stack\n");
			ret = -ENOMEM;
			goto unmap_buffers;
		}

		dev_dbg(pdata->dev, "stack is allocated at %#08x\n",
			stack->paddr);
	}

	return 0;

unmap_buffers:
	unmap_buffers(job_desc);
detach_buffers:
	detach_buffers(job_desc);
err_fd:
	put_unused_fd(job_desc->job.fd);
err_desc:
	kfree(job_desc);
	return ret;
}

static int elcore_job_enqueue(struct elcore_private_data *pdata,
			      void __user *arg)
{
	struct elcore_job job;
	struct elcore_job_desc *job_desc;
	unsigned long flags;
	struct fd fd;
	int ret;

	ret = copy_from_user(&job, arg, sizeof(struct elcore_job));
	if (ret)
		return -EACCES;

	fd = fdget(job.fd);
	if (!fd.file || fd.file->f_op != &elcore_job_fops) {
		ret = -EBADFD;
		goto done;
	}

	job_desc = fd.file->private_data;

	/* TODO: Before adding job to queue we should check that all job
	 * buffers were unmapped to provide memory coherency.
	 */

	spin_lock_irqsave(&job_desc->lock, flags);
	if (job_desc->job.status != ELCORE_JOB_IDLE) {
		spin_unlock_irqrestore(&job_desc->lock, flags);
		ret = -EBUSY;
		goto done;
	}
	job_desc->job.status = ELCORE_JOB_ENQUEUED;
	spin_unlock_irqrestore(&job_desc->lock, flags);

	spin_lock_irqsave(&pdata->lock, flags);
	list_add(&job_desc->job_queue_node, &pdata->job_queue);
	if (list_is_singular(&job_desc->job_queue_node))
		elcore_run_job(job_desc);
	spin_unlock_irqrestore(&pdata->lock, flags);

	dev_dbg(pdata->dev, "elcore job @ %p is added\n", job_desc);

done:
	fdput(fd);
	return ret;
}

static int elcore_job_status(void __user *arg)
{
	struct elcore_job job;
	struct elcore_job_desc *job_desc;
	struct fd fd;
	int ret;

	ret = copy_from_user(&job, arg, sizeof(struct elcore_job));
	if (ret)
		return -EACCES;

	fd = fdget(job.fd);
	if (!fd.file || fd.file->f_op != &elcore_job_fops) {
		ret = -EBADFD;
		goto done;
	}

	job_desc = fd.file->private_data;

	ret = copy_to_user(arg, &job_desc->job, sizeof(struct elcore_job));
	if (ret)
		ret = -EFAULT;

done:
	fdput(fd);
	return ret;
}

static int elcore_job_cancel(struct elcore_private_data *pdata,
			     void __user *arg)
{
	struct elcore_job job;
	struct elcore_job_desc *job_desc;
	unsigned long flags;
	struct fd fd;
	int ret;

	ret = copy_from_user(&job, arg, sizeof(struct elcore_job));
	if (ret)
		return -EACCES;

	fd = fdget(job.fd);
	if (!fd.file || fd.file->f_op != &elcore_job_fops) {
		ret = -EBADFD;
		goto done;
	}

	job_desc = fd.file->private_data;

	spin_lock_irqsave(&job_desc->lock, flags);
	switch (job_desc->job.status) {
	case ELCORE_JOB_IDLE:
		break;
	case ELCORE_JOB_ENQUEUED:
		spin_lock(&pdata->lock);
		list_del(&job_desc->job_queue_node);
		spin_unlock(&pdata->lock);

		job_desc->job.status = ELCORE_JOB_IDLE;
		job_desc->job.rc = ELCORE_JOB_CANCELLED;
		break;
	case ELCORE_JOB_RUNNING:
		/* TODO: Stop this job */
		ret = -EBUSY;
		break;
	}
	spin_unlock_irqrestore(&job_desc->lock, flags);

done:
	fdput(fd);
	return ret;
}

static int elcore_dmachain_setup(struct elcore_private_data *pdata,
				 void __user *arg)
{
	struct elcore_buffer_desc *external = NULL, *chain = NULL;
	struct elcore_buffer_desc *internal[2] = { NULL };
	struct elcore_dmachain c;
	struct vdma_descriptor *vd;
	int ret, odd = 0;

	ret = copy_from_user(&c, arg, sizeof(struct elcore_dmachain));
	if (ret)
		return -EACCES;

	external = elcore_buf_desc_get(pdata->dev, c.external);
	if (IS_ERR(external)) {
		ret = PTR_ERR(external);
		goto done;
	}

	internal[0] = elcore_buf_desc_get(pdata->dev, c.internal[0]);
	if (IS_ERR(internal[0])) {
		ret = PTR_ERR(internal[0]);
		goto done;
	}

	internal[1] = elcore_buf_desc_get(pdata->dev, c.internal[1]);
	if (IS_ERR(internal[1])) {
		ret = PTR_ERR(internal[1]);
		goto done;
	}

	chain = elcore_buf_desc_get(pdata->dev, c.chain);
	if (IS_ERR(chain)) {
		ret = PTR_ERR(chain);
		goto done;
	}

	if (external->buf.type != ELCORE_MEMORY_SYSTEM ||
	    internal[0]->buf.type != ELCORE_MEMORY_VRAM ||
	    internal[1]->buf.type != ELCORE_MEMORY_VRAM) {
		ret = -EINVAL;
		goto done;
	}

	vd = chain->vaddr;

	do {
		void *const next = vd->a_init + chain->vaddr;

		/* Prevent out-of-range access */
		if (next - chain->vaddr > chain->buf.size) {
			ret = -EFAULT;
			break;
		}

		vd->a0e += external->paddr;
		vd->aecur += external->paddr;

		/*
		 * TODO: This is a dirty hack. There should be a general way to
		 * determine dma-specific address of buffer.
		 * TODO: Currently we do not know how VDMA addresses internal
		 * memory.
		 */
		vd->a0i += internal[odd]->paddr;
		vd->aicur += internal[odd]->paddr;
		odd ^= 1;

		if (vd->a_init)
			/* Pointer to the next descriptor */
			vd->a_init += chain->paddr;
		else
			break;

		vd = next;
	} while (true);

done:
	elcore_buf_desc_put(chain);
	elcore_buf_desc_put(internal[1]);
	elcore_buf_desc_put(internal[0]);
	elcore_buf_desc_put(external);

	return ret;
}

static long elcore_ioctl(struct file *file, unsigned int cmd,
			 unsigned long arg)
{
	struct elcore_private_data *pdata =
		(struct elcore_private_data *)file->private_data;
	void __user *const uptr = (void __user *)arg;

	switch (cmd) {
	case ELCIOC_BUF_ALLOC:
		return elcore_buf_alloc(pdata, uptr);
	case ELCIOC_RESOURCE_REQUEST:
		return elcore_resource_request(pdata, uptr);
	case ELCIOC_RESOURCE_RELEASE:
		return elcore_resource_release(pdata, uptr);
	case ELCIOC_JOB_CREATE:
		return elcore_job_create(pdata, uptr);
	case ELCIOC_JOB_ENQUEUE:
		return elcore_job_enqueue(pdata, uptr);
	case ELCIOC_JOB_STATUS:
		return elcore_job_status(uptr);
	case ELCIOC_JOB_CANCEL:
		return elcore_job_cancel(pdata, uptr);
	case ELCIOC_DMACHAIN_SETUP:
		return elcore_dmachain_setup(pdata, uptr);
	}

	return -ENOTTY;
}

static int elcore_open(struct inode *inode, struct file *file)
{
	struct elcore_private_data *pdata;

	pdata = container_of(inode->i_cdev, struct elcore_private_data, cdev);
	file->private_data = pdata;
	return 0;
}

static int elcore_release(struct inode *inode, struct file *file)
{
	return 0;
}

static struct file_operations elcore_fops = {
	.owner = THIS_MODULE,
	.open = elcore_open,
	.release = elcore_release,
	.compat_ioctl = elcore_ioctl,
	.unlocked_ioctl = elcore_ioctl
};

static int elcore_init_codebuf(struct device *dev, struct region *region)
{
	struct device_node *node;
	int rc;

	node = of_parse_phandle(dev->of_node, "codebuf", 0);
	if (!node)
		return -ENOENT;

	rc = of_address_to_resource(node, 0, &region->res);
	if (rc < 0)
		return -ENOENT;

	mutex_init(&region->requested);

	dev_info(dev, "codebuf: %pR\n", &region->res);

	return 0;
}

static int elcore_probe(struct platform_device *pdev)
{
	struct elcore_private_data *pdata;
	struct resource *regs;
	struct device *dev;
	struct regmap *cr;
	struct reset_control *velcore_reset;
	int irq, ret, i;

	regs = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!regs) {
		dev_err(&pdev->dev, "Can not get memory resource\n");
		return -ENOENT;
	}

	pdata = devm_kzalloc(&pdev->dev, sizeof(struct elcore_private_data),
			     GFP_KERNEL);
	if (!pdata)
		return -ENOMEM;

	pdata->dev = &pdev->dev;
	INIT_LIST_HEAD(&pdata->job_queue);
	spin_lock_init(&pdata->lock);

	pdata->regs = devm_ioremap_resource(&pdev->dev, regs);
	if (IS_ERR(pdata->regs))
		return PTR_ERR(pdata->regs);


	irq = platform_get_irq(pdev, 0);
	ret = devm_request_irq(&pdev->dev, irq, elcore_interrupt, IRQF_SHARED,
			       pdev->name, pdata);
	if (ret) {
		dev_err(&pdev->dev, "Can not get interrupt resource\n");
		return -ENOENT;
	}

	pdata->vpool = of_get_named_gen_pool(pdev->dev.of_node, "vram", 0);
	if (!pdata->vpool) {
		dev_err(&pdev->dev, "Can not get VRAM pool\n");
		return -ENOENT;
	}

	ret = elcore_init_codebuf(&pdev->dev, &pdata->codebuf);
	if (ret < 0)
		dev_warn(&pdev->dev, "Can not initialize code buffer\n");

	ret = alloc_chrdev_region(&pdata->dev_num, 0, 1, "elcore");
	if (ret < 0) {
		dev_err(&pdev->dev,
			"Can not allocate region for elcore chrdev\n");
		return ret;
	}

	pdata->class = class_create(THIS_MODULE, "elcore");
	if (IS_ERR(pdata->class)) {
		dev_err(&pdev->dev, "Can not create class for elcore\n");
		ret = PTR_ERR(pdata->class);
		goto err1;
	}

	dev = device_create(pdata->class, NULL, pdata->dev_num, NULL, "elc");
	if (IS_ERR(dev)) {
		dev_err(&pdev->dev, "Can not create elc device\n");
		ret = PTR_ERR(dev);
		goto err2;
	}

	cdev_init(&pdata->cdev, &elcore_fops);
	ret = cdev_add(&pdata->cdev, pdata->dev_num, 1);
	if (ret < 0) {
		dev_err(&pdev->dev, "Can not add elcore cdev\n");
		goto err3;
	}

	platform_set_drvdata(pdev, pdata);

	/*
	 * TODO: Consider following register tuning as temporary
	 */

	/* This code is designed for PLL1 set to 2000 MHz */
	cr = syscon_regmap_lookup_by_phandle(pdev->dev.of_node,
					     "img,cr-top-secure");
	if (IS_ERR(cr)) {
		ret = PTR_ERR(cr);
		dev_err(&pdev->dev, "Cannot look up cr-top-secure registers\n");
		goto err_cdev_del;
	}

	/* Turn off FFT clock */
	regmap_write(cr, 0xe8, 0x00);

	/* SYS3 = PLL1 / sys3div */
	regmap_write(cr, 0xf0, ((sys3div - 1) << 6) | 0x1);
	regmap_write(cr, 0xf0, ((sys3div - 1) << 6) | 0x9);

	/* CIP3 = PLL1 / cip3div */
	regmap_write(cr, 0x140, ((cip3div - 1) << 7) | 0x1);
	regmap_write(cr, 0x140, ((cip3div - 1) << 7) | 0x11);

	/* TODO: proper power management */
	/* This code is written as workaround for rf#5422 */
	velcore_reset = devm_reset_control_get(&pdev->dev, NULL);
	if (IS_ERR(velcore_reset)) {
		dev_info(&pdev->dev, "Cannot get reset control\n");
		dev_info(&pdev->dev, "Will continue without VDMA workaround\n");
	} else
		for (i = 0; i < 2; i++) {
			elcore_writel(pdata, CTR_CLK, 0x1FF);
			reset_control_reset(velcore_reset);
		}

	/* Enable clocks to the first core and HWSGBM */
	elcore_writel(pdata, CTR_CLK, 0x101);

	/* unmask stop interrupt for CORE0 and all IRQs for HWSGBM */
	elcore_writel(pdata, MASKR_ELCORE, MASKR_CORE0_STOP | HWSGBM_INTS);
	/*
	 * Switch on all IRQs in HWSGBM IRQ mask register.
	 * TODO: untether them from DSP driver.
	 */
	elcore_writel(pdata, MASKR_ACC, 0x1f);

	dev_dbg(&pdev->dev, "ELVEES Elcore40 @ 0x%p\n", pdata->regs);

	return 0;
err_cdev_del:
	cdev_del(&pdata->cdev);
err3:
	device_destroy(pdata->class, pdata->dev_num);
err2:
	class_destroy(pdata->class);
err1:
	unregister_chrdev_region(pdata->dev_num, 1);

	dev_err(&pdev->dev, "Elcore40 init failed, error %d\n", ret);
	return ret;
}

static int elcore_remove(struct platform_device *pdev)
{
	struct elcore_private_data *pdata = platform_get_drvdata(pdev);

	device_destroy(pdata->class, pdata->dev_num);
	class_destroy(pdata->class);
	unregister_chrdev_region(pdata->dev_num, 1);

	/* TODO: Stop all jobs on remove */

	/* mask stop interrupt for CORE0 */
	elcore_writel(pdata, MASKR_ELCORE, 0);

	return 0;
}

#ifdef CONFIG_OF
static struct of_device_id elcore_dt_ids[] = {
	{ .compatible = "elvees,elcore40" },
	{}
};
MODULE_DEVICE_TABLE(of, elcore_dt_ids);
#endif

static struct platform_driver elcore_driver = {
	.driver = {
		.name = "elcore40",
		.of_match_table = of_match_ptr(elcore_dt_ids),
	},
	.probe = elcore_probe,
	.remove = elcore_remove,
};

module_platform_driver(elcore_driver);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("ELVEES Elcore40 driver");
MODULE_AUTHOR("Dmitriy Zagrebin <dzagrebin@elvees.com>");
