/*
 * HWSGBM driver
 *
 * Copyright 2016 ELVEES NeoTek JSC
 * Author: Oleg Kitain <okitain@elvees.com>
 */

#define pr_fmt(fmt) KBUILD_MODNAME ": " fmt

#include <linux/genalloc.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/spinlock.h>
#include <linux/device.h>
#include <linux/slab.h>
#include <linux/of.h>
#include <linux/types.h>
#include <linux/irqreturn.h>
#include <linux/interrupt.h>
#include <linux/clk.h>

#include <media/v4l2-mem2mem.h>
#include <media/v4l2-device.h>
#include <media/v4l2-ioctl.h>
#include <media/v4l2-ctrls.h>
#include <media/v4l2-event.h>
#include <media/videobuf2-dma-contig.h>
#include "hwsgbm-driver.h"
#include "hwsgbm-vram.h"
#include "hwsgbm-vdma.h"
#include "hwsgbm.h"

MODULE_DESCRIPTION("HWSGBM driver");
MODULE_LICENSE("GPL");
MODULE_VERSION("0.1");
MODULE_AUTHOR("Oleg Kitain <okitain@elvees.com>");

#define MODULE_NAME "hwsgbm"

static unsigned int use_vdma = 1;
module_param(use_vdma, uint, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(use_vdma, "use vdma?");

static struct v4l2_fmtdesc output_formats[] = {
	{
		.index = 0,
		.type  = V4L2_BUF_TYPE_VIDEO_OUTPUT,
		.description = "NV21",
		.pixelformat = V4L2_PIX_FMT_NV21
	},
	{
		.index = 1,
		.type  = V4L2_BUF_TYPE_VIDEO_OUTPUT,
		.description = "NV12",
		.pixelformat = V4L2_PIX_FMT_NV12
	},
	{
		.index = 2,
		.type  = V4L2_BUF_TYPE_VIDEO_OUTPUT,
		.description = "YUV420",
		.pixelformat = V4L2_PIX_FMT_YUV420
	},
	{
		.index = 3,
		.type  = V4L2_BUF_TYPE_VIDEO_OUTPUT,
		.description = "YVU420",
		.pixelformat = V4L2_PIX_FMT_YVU420
	},
	{
		.index = 4,
		.type  = V4L2_BUF_TYPE_VIDEO_OUTPUT,
		.description = "Grayscale 8bit",
		.pixelformat = V4L2_PIX_FMT_GREY
	}
};

static struct v4l2_fmtdesc capture_formats[] = {
	{
		.index = 0,
		.type  = V4L2_BUF_TYPE_VIDEO_CAPTURE,
		.description = "Grayscale 8bit",
		.pixelformat = V4L2_PIX_FMT_GREY
	}
};

static int ctx_is_active(struct hwsgbm_ctx *ctx)
{
	return ctx->state == HWSGBM_CTX_ACTIVE;
}

/* Privilege function design borrowed from UVC */
static int ctx_activate(struct hwsgbm_ctx *ctx)
{
	/* Always succeed if the ctx is already privileged. */
	if (ctx_is_active(ctx))
		return 0;

	/* Check if any other ctx is privileged */
	if (atomic_inc_return(&ctx->drv_priv->active) != 1) {
		atomic_dec(&ctx->drv_priv->active);
		return -EBUSY;
	}

	ctx->state = HWSGBM_CTX_ACTIVE;
	return 0;
}

static void ctx_deactivate(struct hwsgbm_ctx *ctx)
{
	if (ctx_is_active(ctx))
		atomic_dec(&ctx->drv_priv->active);

	ctx->state = HWSGBM_CTX_PASSIVE;
}

static int is_any_streaming(struct hwsgbm_priv *drv_priv)
{
	return atomic_read(&drv_priv->active);
}

static struct vb2_buffer *v4l2_m2m_next_src_buf_by_field(
		struct v4l2_m2m_ctx *m2m_ctx,
		unsigned int field_type)
{
	struct v4l2_m2m_buffer *b;
	unsigned long flags;
	struct v4l2_m2m_queue_ctx *q_ctx = &(m2m_ctx->out_q_ctx);

	spin_lock_irqsave(&q_ctx->rdy_spinlock, flags);

	if (list_empty(&q_ctx->rdy_queue)) {
		spin_unlock_irqrestore(&q_ctx->rdy_spinlock, flags);
		return NULL;
	}
	b = list_first_entry(&q_ctx->rdy_queue, struct v4l2_m2m_buffer, list);
	while (b->vb.v4l2_buf.field != field_type) {
		b = list_next_entry(b, list);
		if (&(b->list) == &q_ctx->rdy_queue) {
			spin_unlock_irqrestore(&q_ctx->rdy_spinlock, flags);
			return NULL;
		}
	}

	spin_unlock_irqrestore(&q_ctx->rdy_spinlock, flags);
	return &b->vb;
}

static struct vb2_buffer *v4l2_m2m_src_buf_remove_by_field(
		struct v4l2_m2m_ctx *m2m_ctx,
		unsigned int field_type)
{
	struct v4l2_m2m_buffer *b;
	unsigned long flags;
	struct v4l2_m2m_queue_ctx *q_ctx = &(m2m_ctx->out_q_ctx);

	spin_lock_irqsave(&q_ctx->rdy_spinlock, flags);
	if (list_empty(&q_ctx->rdy_queue)) {
		spin_unlock_irqrestore(&q_ctx->rdy_spinlock, flags);
		return NULL;
	}
	b = list_first_entry(&q_ctx->rdy_queue, struct v4l2_m2m_buffer, list);
	while (b->vb.v4l2_buf.field != field_type) {
		b = list_next_entry(b, list);
		if (&(b->list) == &q_ctx->rdy_queue) {
			spin_unlock_irqrestore(&q_ctx->rdy_spinlock, flags);
			return NULL;
		}
	}
	list_del(&b->list);
	q_ctx->num_rdy--;
	spin_unlock_irqrestore(&q_ctx->rdy_spinlock, flags);

	return &b->vb;
}

static void hwsgbm_calc_params(struct hwsgbm_ctx *ctx)
{
	/*
	 * Normally disp_width would be equal to (width - nod - 6), but we
	 * reduce disparity map width by 1 to make DMAP_HSIZE even, which
	 * is required per HWSGBM spec.
	 */
	ctx->disp_width = ctx->width - ctx->nod - 6 - 1;
	ctx->disp_height = ctx->height - 4;
	/* TODO: accept stride from user */
	ctx->stride = round_up(ctx->width, 16);
	ctx->disp_stride = round_up(ctx->disp_width, 16);
}

static int ctx_s_ctrl(struct v4l2_ctrl *ctrl)
{
	u32 reg;
	struct hwsgbm_ctx *ctx = ctrl->priv;

	if (ctx_is_active(ctx))
		return -EBUSY;

	switch (ctrl->id) {
	case V4L2_CID_HWSGBM_NOD:
		/*
		 * the size of resulting image is dependent on NoD, so NoD
		 * should not be changed when cap buffers are already allocated
		 */
		if (vb2_is_busy(v4l2_m2m_get_vq(ctx->fh.m2m_ctx,
						V4L2_BUF_TYPE_VIDEO_CAPTURE)))
			return -EBUSY;
		ctx->nod = ctrl->val;
		hwsgbm_calc_params(ctx);
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

static int hwsgbm_s_ctrl(struct v4l2_ctrl *ctrl)
{
	u32 reg;
	struct hwsgbm_priv *drv_priv = ctrl->priv;

	if (is_any_streaming(drv_priv))
		return -EBUSY;

	switch (ctrl->id) {
	case V4L2_CID_HWSGBM_ACCUMULATION:
		reg = hwsgbm_read(drv_priv, SGBM_CONFIG);
		reg &= ~ACC_DIRECT_MASK;
		reg |= ctrl->val & ACC_DIRECT_MASK;
		hwsgbm_write(reg, drv_priv, SGBM_CONFIG);
		drv_priv->params.acc_direct = ctrl->val;
		break;
	case V4L2_CID_HWSGBM_PEAK_FILTER:
		reg = hwsgbm_read(drv_priv, SGBM_CONFIG);
		reg &= ~PEAK_EN;
		reg |= ctrl->val ? PEAK_EN : 0;
		hwsgbm_write(reg, drv_priv, SGBM_CONFIG);
		drv_priv->params.peak_en = ctrl->val;
		break;
	case V4L2_CID_HWSGBM_MAX_PEAK_SIZE:
		reg = hwsgbm_read(drv_priv, SGBM_PEAK_PARAM);
		reg &= ~MAXPEAK_MASK;
		reg |= ctrl->val & MAXPEAK_MASK;
		hwsgbm_write(reg, drv_priv, SGBM_PEAK_PARAM);
		break;
	case V4L2_CID_HWSGBM_INTERPOLATION:
		reg = hwsgbm_read(drv_priv, SGBM_CONFIG);
		reg &= ~INTERP_EN;
		reg |= ctrl->val ? INTERP_EN : 0;
		hwsgbm_write(reg, drv_priv, SGBM_CONFIG);
		drv_priv->params.interpolate_en = ctrl->val;
		break;
	case V4L2_CID_HWSGBM_INTERPOLATION_RADIUS:
		reg = hwsgbm_read(drv_priv, SGBM_INTERP_PARAM);
		reg &= ~INTERP_RADIUS_MASK;
		reg |= ctrl->val & INTERP_RADIUS_MASK;
		hwsgbm_write(reg, drv_priv, SGBM_INTERP_PARAM);
		break;
	case V4L2_CID_HWSGBM_ANISOTROPIC_FILTER:
		reg = hwsgbm_read(drv_priv, SGBM_INTERP_PARAM);
		reg &= ~EDGE_EN;
		reg |= ctrl->val ? EDGE_EN : 0;
		hwsgbm_write(reg, drv_priv, SGBM_INTERP_PARAM);
		break;
	case V4L2_CID_HWSGBM_ISLAND_FILTER:
		reg = hwsgbm_read(drv_priv, SGBM_INTERP_PARAM);
		reg &= ~ISLAND_EN;
		reg |= ctrl->val ? ISLAND_EN : 0;
		hwsgbm_write(reg, drv_priv, SGBM_INTERP_PARAM);
		break;
	case V4L2_CID_HWSGBM_ISOTROPIC_FILTER:
		reg = hwsgbm_read(drv_priv, SGBM_INTERP_PARAM);
		reg &= ~NUM_VALID_MASK;
		reg |= (ctrl->val << NUM_VALID_SHIFT) & NUM_VALID_MASK;
		hwsgbm_write(reg, drv_priv, SGBM_INTERP_PARAM);
		break;
	case V4L2_CID_HWSGBM_MEDIAN_FILTER:
		reg = hwsgbm_read(drv_priv, SGBM_CONFIG);
		reg &= ~MEDIAN_EN;
		reg |= ctrl->val ? MEDIAN_EN : 0;
		hwsgbm_write(reg, drv_priv, SGBM_CONFIG);
		drv_priv->params.median_en = ctrl->val;
		break;
	case V4L2_CID_HWSGBM_MEDIAN_FILTER_STAGES:
		reg = hwsgbm_read(drv_priv, SGBM_MEDIAN_PARAM);
		reg &= ~STAGES_NUM_MASK;
		reg |= ctrl->val & STAGES_NUM_MASK;
		hwsgbm_write(reg, drv_priv, SGBM_MEDIAN_PARAM);
		break;
	case V4L2_CID_HWSGBM_MEDIAN_REPLACE_INVALID_LOW:
		reg = ctrl->p_new.p_u32[0];
		hwsgbm_write(reg, drv_priv, SGBM_MEDIAN_REP1);
		break;
	case V4L2_CID_HWSGBM_MEDIAN_REPLACE_INVALID_HIGH:
		reg = ctrl->p_new.p_u32[0];
		hwsgbm_write(reg, drv_priv, SGBM_MEDIAN_REP2);
		break;
	case V4L2_CID_HWSGBM_SOBEL_THRESHOLD:
		reg = ctrl->val;
		hwsgbm_write(reg, drv_priv, SGBM_FILT_PARAM);
		break;
	case V4L2_CID_HWSGBM_PENALTY_FOR_SINGLE:
		reg = hwsgbm_read(drv_priv, SGBM_PENALTY);
		reg &= ~PENALTY1_MASK;
		reg |= ctrl->val & PENALTY1_MASK;
		hwsgbm_write(reg, drv_priv, SGBM_PENALTY);
		break;
	case V4L2_CID_HWSGBM_PENALTY_FOR_MULTIPLE:
		reg = hwsgbm_read(drv_priv, SGBM_PENALTY);
		reg &= ~PENALTY2_MASK;
		reg |= (ctrl->val << PENALTY2_SHIFT) & PENALTY2_MASK;
		hwsgbm_write(reg, drv_priv, SGBM_PENALTY);
		break;
	case V4L2_CID_HWSGBM_MAX_LR_DIFFERENCE:
		reg = hwsgbm_read(drv_priv, SGBM_CONFIG);
		reg &= ~MAX_LR_DIFF_MASK;
		reg |= (ctrl->val << MAX_LR_DIFF_SHIFT) & MAX_LR_DIFF_MASK;
		hwsgbm_write(reg, drv_priv, SGBM_CONFIG);
		break;
	case V4L2_CID_HWSGBM_MEDIAN_REPLACE_VALID_LOW:
		reg = ctrl->p_new.p_u32[0];
		hwsgbm_write(reg, drv_priv, SGBM_INTUSE_MEDIAN_R1);
		break;
	case V4L2_CID_HWSGBM_MEDIAN_REPLACE_VALID_HIGH:
		reg = ctrl->p_new.p_u32[0];
		hwsgbm_write(reg, drv_priv, SGBM_INTUSE_MEDIAN_R2);
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

static int ctx_try_ctrl(struct v4l2_ctrl *ctrl)
{
	struct hwsgbm_ctx *ctx = ctrl->priv;

	switch (ctrl->id) {
	case V4L2_CID_HWSGBM_NOD:
		if (ctrl->val > 127 && ctx->drv_priv->params.acc_direct != 1) {
			v4l2_printk(KERN_NOTICE, &ctx->drv_priv->v4l2_dev,
			    "Cannot use NoD>127 with LR-check\n");
			return -EINVAL;
		}
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

static int hwsgbm_try_ctrl(struct v4l2_ctrl *ctrl)
{
	switch (ctrl->id) {
	case V4L2_CID_HWSGBM_ACCUMULATION:
	case V4L2_CID_HWSGBM_PEAK_FILTER:
	case V4L2_CID_HWSGBM_MAX_PEAK_SIZE:
	case V4L2_CID_HWSGBM_INTERPOLATION:
	case V4L2_CID_HWSGBM_INTERPOLATION_RADIUS:
	case V4L2_CID_HWSGBM_ANISOTROPIC_FILTER:
	case V4L2_CID_HWSGBM_ISLAND_FILTER:
	case V4L2_CID_HWSGBM_ISOTROPIC_FILTER:
	case V4L2_CID_HWSGBM_MEDIAN_FILTER:
	case V4L2_CID_HWSGBM_MEDIAN_FILTER_STAGES:
	case V4L2_CID_HWSGBM_MEDIAN_REPLACE_INVALID_LOW:
	case V4L2_CID_HWSGBM_MEDIAN_REPLACE_INVALID_HIGH:
	case V4L2_CID_HWSGBM_SOBEL_THRESHOLD:
	case V4L2_CID_HWSGBM_PENALTY_FOR_SINGLE:
	case V4L2_CID_HWSGBM_PENALTY_FOR_MULTIPLE:
	case V4L2_CID_HWSGBM_MAX_LR_DIFFERENCE:
	case V4L2_CID_HWSGBM_MEDIAN_REPLACE_VALID_LOW:
	case V4L2_CID_HWSGBM_MEDIAN_REPLACE_VALID_HIGH:
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

static const struct v4l2_ctrl_ops ctrl_drv_ops = {
	.s_ctrl = hwsgbm_s_ctrl,
	.try_ctrl = hwsgbm_try_ctrl,
};

static const struct v4l2_ctrl_ops ctrl_ctx_ops = {
	.s_ctrl = ctx_s_ctrl,
	.try_ctrl = ctx_try_ctrl,
};

static const char * const acc_direct_strings[] = {
	"1D, LR-check",
	"2D",
	"skip",
	"2D, LR-check",
	NULL,
};

static const struct v4l2_ctrl_config ctrl_cfg[] = {
	{
		.ops   = &ctrl_ctx_ops,
		.id    = V4L2_CID_HWSGBM_NOD,
		.name  = "nod",
		.type  = V4L2_CTRL_TYPE_INTEGER,
		.min   = 15,
		.max   = 239,
		.step  = 16,
		.def   = 127,
	},
	{
		.ops   = &ctrl_drv_ops,
		.id    = V4L2_CID_HWSGBM_ACCUMULATION,
		.name  = "accumulation direction",
		.type  = V4L2_CTRL_TYPE_MENU,
		.min   = 0,
		.max   = 3,
		.def   = 3,
		.qmenu = acc_direct_strings,
		.menu_skip_mask = BIT(2),
	},
	{
		.ops   = &ctrl_drv_ops,
		.id    = V4L2_CID_HWSGBM_PEAK_FILTER,
		.name  = "peak filter",
		.type  = V4L2_CTRL_TYPE_BOOLEAN,
		.min   = 0,
		.max   = 1,
		.step  = 1,
		.def   = 1,
	},
	{
		.ops   = &ctrl_drv_ops,
		.id    = V4L2_CID_HWSGBM_MAX_PEAK_SIZE,
		.name  = "max peak size",
		.type  = V4L2_CTRL_TYPE_INTEGER,
		.min   = 0,
		.max   = 63,
		.step  = 1,
		.def   = 15,
	},
	{
		.ops   = &ctrl_drv_ops,
		.id    = V4L2_CID_HWSGBM_INTERPOLATION,
		.name  = "interpolation",
		.type  = V4L2_CTRL_TYPE_BOOLEAN,
		.min   = 0,
		.max   = 1,
		.step  = 1,
		.def   = 1,
	},
	{
		.ops   = &ctrl_drv_ops,
		.id    = V4L2_CID_HWSGBM_INTERPOLATION_RADIUS,
		.name  = "interpolation radius",
		.type  = V4L2_CTRL_TYPE_INTEGER,
		.min   = 1,
		.max   = 8,
		.step  = 1,
		.def   = 8,
	},
	{
		.ops   = &ctrl_drv_ops,
		.id    = V4L2_CID_HWSGBM_ANISOTROPIC_FILTER,
		.name  = "anisotropic filter",
		.type  = V4L2_CTRL_TYPE_BOOLEAN,
		.min   = 0,
		.max   = 1,
		.step  = 1,
		.def   = 0,
	},
	{
		.ops   = &ctrl_drv_ops,
		.id    = V4L2_CID_HWSGBM_ISLAND_FILTER,
		.name  = "island filter",
		.type  = V4L2_CTRL_TYPE_BOOLEAN,
		.min   = 0,
		.max   = 1,
		.step  = 1,
		.def   = 0,
	},
	{
		.ops   = &ctrl_drv_ops,
		.id    = V4L2_CID_HWSGBM_ISOTROPIC_FILTER,
		.name  = "isotropic filter",
		.type  = V4L2_CTRL_TYPE_INTEGER,
		.min   = 0,
		.max   = 8,
		.step  = 1,
		.def   = 3,
	},
	{
		.ops   = &ctrl_drv_ops,
		.id    = V4L2_CID_HWSGBM_MEDIAN_FILTER,
		.name  = "median filter",
		.type  = V4L2_CTRL_TYPE_BOOLEAN,
		.min   = 0,
		.max   = 1,
		.step  = 1,
		.def   = 1,
	},
	{
		.ops   = &ctrl_drv_ops,
		.id    = V4L2_CID_HWSGBM_MEDIAN_FILTER_STAGES,
		.name  = "median filter stages",
		.type  = V4L2_CTRL_TYPE_INTEGER,
		.min   = 0,
		.max   = 8,
		.step  = 1,
		.def   = 8,
	},
	{
		.ops   = &ctrl_drv_ops,
		.id    = V4L2_CID_HWSGBM_MEDIAN_REPLACE_INVALID_LOW,
		.name  = "median replace invalid low",
		.type  = V4L2_CTRL_TYPE_U32,
		.min   = 0,
		.max   = 0xFFFFFFFF,
		.step  = 1,
		.def   = 0,
		.dims[0] = 1,
	},
	{
		.ops   = &ctrl_drv_ops,
		.id    = V4L2_CID_HWSGBM_MEDIAN_REPLACE_INVALID_HIGH,
		.name  = "median replace invalid high",
		.type  = V4L2_CTRL_TYPE_U32,
		.min   = 0,
		.max   = 0xFFFFFFFF,
		.step  = 1,
		.def   = 0,
		.dims[0] = 1,
	},
	{
		.ops   = &ctrl_drv_ops,
		.id    = V4L2_CID_HWSGBM_SOBEL_THRESHOLD,
		.name  = "sobel threshold",
		.type  = V4L2_CTRL_TYPE_INTEGER,
		.min   = 0,
		.max   = 127,
		.step  = 1,
		.def   = 60,
	},
	{
		.ops   = &ctrl_drv_ops,
		.id    = V4L2_CID_HWSGBM_PENALTY_FOR_SINGLE,
		.name  = "penalty for single",
		.type  = V4L2_CTRL_TYPE_INTEGER,
		.min   = 0,
		.max   = 1275,
		.step  = 1,
		.def   = 96,
	},
	{
		.ops   = &ctrl_drv_ops,
		.id    = V4L2_CID_HWSGBM_PENALTY_FOR_MULTIPLE,
		.name  = "penalty for multiple",
		.type  = V4L2_CTRL_TYPE_INTEGER,
		.min   = 0,
		.max   = 1275,
		.step  = 1,
		.def   = 520,
	},
	{
		.ops   = &ctrl_drv_ops,
		.id    = V4L2_CID_HWSGBM_MAX_LR_DIFFERENCE,
		.name  = "max lr difference",
		.type  = V4L2_CTRL_TYPE_INTEGER,
		.min   = 0,
		.max   = 255,
		.step  = 1,
		.def   = 50,
	},
	{
		.ops   = &ctrl_drv_ops,
		.id    = V4L2_CID_HWSGBM_MEDIAN_REPLACE_VALID_LOW,
		.name  = "median replace valid low",
		.type  = V4L2_CTRL_TYPE_U32,
		.min   = 0,
		.max   = 0xFFFFFFFF,
		.step  = 1,
		.def   = 0,
		.dims[0] = 1,
	},
	{
		.ops   = &ctrl_drv_ops,
		.id    = V4L2_CID_HWSGBM_MEDIAN_REPLACE_VALID_HIGH,
		.name  = "median replace valid high",
		.type  = V4L2_CTRL_TYPE_U32,
		.min   = 0,
		.max   = 0xFFFFFFFF,
		.step  = 1,
		.def   = 0,
		.dims[0] = 1,
	},
};

static int hwsgbm_ctrl_init(struct hwsgbm_priv *drv_priv)
{
	int i;
	struct v4l2_ctrl *tmp_ctrl;

	v4l2_ctrl_handler_init(&drv_priv->hdl, 20);
	/* We skip the first ctrl, and initialize it in ctx */
	for (i = 1; i < ARRAY_SIZE(ctrl_cfg); i++) {
		tmp_ctrl = v4l2_ctrl_new_custom(&drv_priv->hdl, &ctrl_cfg[i],
						drv_priv);
		if (!tmp_ctrl) {
			v4l2_err(&drv_priv->v4l2_dev,
				 "Failed to create control: 0x%x\n",
				 ctrl_cfg[i].id);
			v4l2_ctrl_handler_free(&drv_priv->hdl);
			return drv_priv->hdl.error;
		}
	}
	v4l2_ctrl_handler_setup(&drv_priv->hdl);
	drv_priv->vfd.ctrl_handler = &drv_priv->hdl;
	return 0;
};

static void hwsgbm_writeconf(struct hwsgbm_ctx *ctx)
{
	struct hwsgbm_priv *drv_priv = ctx->drv_priv;
	u32 reg_config = 0;
	u32 dmap_param = 0;
	u32 eventmap = 0;
	u32 eventmask = 0;
	int i;

	reg_config = hwsgbm_read(drv_priv, SGBM_CONFIG);

	reg_config &= ~NOD_MASK;
	reg_config |= (ctx->nod << NOD_SHIFT) & NOD_MASK;
	hwsgbm_write(reg_config, drv_priv, SGBM_CONFIG);
	dmap_param |= ctx->disp_width & DMAP_HSIZE_MASK;
	dmap_param |= (ctx->disp_height << DMAP_VSIZE_SHIFT)
		& DMAP_VSIZE_MASK;
	hwsgbm_write(dmap_param, drv_priv, SGBM_DMAP_PARAM);
	hwsgbm_write(SGBM_SL0|SGBM_SL1|SGBM_SR0|SGBM_SR1
		     |SGBM_R1|SGBM_R0|SGBM_EFRAME,
		     drv_priv, SGBM_IRQMASK);

	for (i = 0; i < VDMA_CHAN_NUM; i++) {
		eventmap |= drv_priv->vdma_chmap[i] << EVT_SHIFT(i);
		eventmask |= 1 << EVT_SHIFT(i);
	}
	if (use_vdma) {
		hwsgbm_write(eventmap, drv_priv, SGBM_EVTMAP);
		hwsgbm_write(0, drv_priv, SGBM_EVTMASK);
	} else
		hwsgbm_write(eventmask, drv_priv, SGBM_EVTMASK);
}

static void srcbuf_read(struct hwsgbm_ctx *ctx,
		uint32_t regbit, uint32_t *lines,
		void *vaddr_ptr, int32_t vram_ofst, int odd)
{
	struct hwsgbm_priv *drv_priv = ctx->drv_priv;
	uint32_t status_reg = hwsgbm_read(drv_priv, SGBM_STATUS);
	if ((odd == *lines / BUF_SRC_VSIZE % 2)
	    && (*lines < ctx->height)
	    && (status_reg & regbit || *lines <= BUF_SRC_VSIZE)) {
		hwsgbm_write(regbit, drv_priv, SGBM_STATUS);
		/* TODO: rework for variable stride */
		memcpy(drv_priv->vram_virt + vram_ofst,
		       vaddr_ptr + ctx->stride * *lines,
		       ctx->stride * BUF_SRC_VSIZE);
		hwsgbm_write(regbit, drv_priv, SGBM_CONTROL);
		*lines += BUF_SRC_VSIZE;
	}
}

static int dstbuf_write(struct hwsgbm_ctx *ctx,
		uint32_t regbit, uint32_t *lines,
		void *vaddr_ptr, int32_t vram_ofst, int odd)
{
	struct hwsgbm_priv *drv_priv = ctx->drv_priv;
	uint32_t status_reg = hwsgbm_read(drv_priv, SGBM_STATUS);

	if (*lines >= ctx->disp_height)
		return 1;
	if ((odd == *lines / BUF_DISP_VSIZE % 2)
	     && (status_reg & regbit)) {
		hwsgbm_write(regbit, drv_priv, SGBM_STATUS);
		/* TODO: rework for variable stride */
		memcpy(vaddr_ptr + ctx->disp_stride * *lines,
		       drv_priv->vram_virt + vram_ofst,
		       ctx->disp_stride * BUF_DISP_VSIZE);
		hwsgbm_write(regbit, drv_priv, SGBM_CONTROL);
		*lines += BUF_DISP_VSIZE;
	}
	return 0;
}

static irqreturn_t hwsgbm_irq_handler(int irq, void *ptr)
{
	struct vb2_buffer *srcl, *srcr, *dst;
	struct hwsgbm_priv *drv_priv = ptr;
	struct hwsgbm_ctx *ctx;
	u32 status, reg_config, buf_status;
	unsigned long flags;

	status = hwsgbm_read(drv_priv, SGBM_STATUS);
	status &= (SGBM_EFRAME | SGBM_ERROR);
	if (!status)
		return IRQ_NONE;

	ctx = v4l2_m2m_get_curr_priv(drv_priv->m2m_dev);

	if (ctx == NULL) {
		pr_err("Instance released before the end of transaction\n");
		return IRQ_HANDLED;
	}

	if (use_vdma)
		hwsgbm_vdma_done(drv_priv);
	srcl = v4l2_m2m_src_buf_remove_by_field(ctx->fh.m2m_ctx,
						V4L2_FIELD_LEFT);
	srcr = v4l2_m2m_src_buf_remove_by_field(ctx->fh.m2m_ctx,
						V4L2_FIELD_RIGHT);
	dst = v4l2_m2m_dst_buf_remove(ctx->fh.m2m_ctx);

	hwsgbm_write(status, drv_priv, SGBM_STATUS);

	reg_config = hwsgbm_read(drv_priv, SGBM_CONFIG);
	hwsgbm_write(reg_config & ~SGBM_EN, drv_priv, SGBM_CONFIG);

	if (status & SGBM_ERROR)
		buf_status = VB2_BUF_STATE_ERROR;
	else {
		buf_status = VB2_BUF_STATE_DONE;
		vb2_set_plane_payload(dst, 0, ctx->disp_stride
					      * ctx->disp_height);
	}
	spin_lock_irqsave(&drv_priv->irqlock, flags);
	v4l2_m2m_buf_done(srcl, buf_status);
	v4l2_m2m_buf_done(srcr, buf_status);
	v4l2_m2m_buf_done(dst, buf_status);
	spin_unlock_irqrestore(&drv_priv->irqlock, flags);

	v4l2_m2m_job_finish(drv_priv->m2m_dev, ctx->fh.m2m_ctx);
	return IRQ_HANDLED;
}


static void hwsgbm_run(void *priv)
{
	struct hwsgbm_ctx *ctx = priv;
	struct hwsgbm_priv *drv_priv = ctx->drv_priv;
	struct vb2_buffer *srcl, *srcr, *dst;
	uint32_t reg_config, reg_irqmask;
	int lines_l = 0, lines_r = 0, lines_dst = 0;
	void *dst_ptr, *srcl_ptr, *srcr_ptr;

	srcl = v4l2_m2m_next_src_buf_by_field(ctx->fh.m2m_ctx,
						V4L2_FIELD_LEFT);
	srcr = v4l2_m2m_next_src_buf_by_field(ctx->fh.m2m_ctx,
						V4L2_FIELD_RIGHT);
	dst = v4l2_m2m_next_dst_buf(ctx->fh.m2m_ctx);

	dst_ptr = vb2_plane_vaddr(dst, 0);
	srcl_ptr = vb2_plane_vaddr(srcl, 0);
	srcr_ptr = vb2_plane_vaddr(srcr, 0);
	if (dst_ptr == NULL || srcl_ptr == NULL || srcr_ptr == NULL) {
		WARN_ON(1);
		return;
	}
	if (use_vdma) {
		dma_addr_t srcl_phys, srcr_phys, dst_phys;

		srcl_phys = vb2_dma_contig_plane_dma_addr(srcl, 0);
		srcr_phys = vb2_dma_contig_plane_dma_addr(srcr, 0);
		dst_phys = vb2_dma_contig_plane_dma_addr(dst, 0);
		hwsgbm_conf_vdma(ctx, srcl_phys, srcr_phys, dst_phys);
	}

	/* We should clear buffer status fields if any of them are still set
	 * from a previous run */

	hwsgbm_write(SGBM_ALL_BUFS, drv_priv, SGBM_STATUS);

	reg_config = hwsgbm_read(drv_priv, SGBM_CONFIG);
	hwsgbm_write(reg_config | SGBM_EN, drv_priv, SGBM_CONFIG);

	if (!use_vdma) {
		hwsgbm_write(SGBM_R0 | SGBM_R1, drv_priv, SGBM_CONTROL);
		while (1) {
			srcbuf_read(ctx, SGBM_SL0, &lines_l,
				srcl_ptr, drv_priv->srcl0_ofst, 0);
			srcbuf_read(ctx, SGBM_SL1, &lines_l,
				srcl_ptr, drv_priv->srcl1_ofst, 1);
			srcbuf_read(ctx, SGBM_SR0, &lines_r,
				srcr_ptr, drv_priv->srcr0_ofst, 0);
			srcbuf_read(ctx, SGBM_SR1, &lines_r,
				srcr_ptr, drv_priv->srcr1_ofst, 1);

			if (dstbuf_write(ctx, SGBM_R0, &lines_dst,
					 dst_ptr, drv_priv->disp0_ofst, 0))
				break;
			if (dstbuf_write(ctx, SGBM_R1, &lines_dst,
					 dst_ptr, drv_priv->disp1_ofst, 1))
				break;
		}
		reg_irqmask = hwsgbm_read(drv_priv, SGBM_IRQMASK);
		hwsgbm_write(reg_irqmask & ~SGBM_EFRAME,
			     drv_priv, SGBM_IRQMASK);
	} else {
		hwsgbm_vdma_run(drv_priv);
	}
};

static int hwsgbm_ready(void *priv)
{
	struct hwsgbm_ctx *ctx = priv;
	struct vb2_buffer *srcl, *srcr;

	srcl = v4l2_m2m_next_src_buf_by_field(ctx->fh.m2m_ctx,
				    V4L2_FIELD_LEFT);
	srcr = v4l2_m2m_next_src_buf_by_field(ctx->fh.m2m_ctx,
				    V4L2_FIELD_RIGHT);
	if (!srcl || !srcr)
		return 0;
	else
		return 1;
}

static void hwsgbm_abort(void *priv)
{
	/* TODO: abort using hardware and a tasklet */
}


static const struct v4l2_m2m_ops hwsgbm_m2m_ops = {
	.device_run = hwsgbm_run,
	.job_ready  = hwsgbm_ready,
	.job_abort  = hwsgbm_abort,
};

static int hwsgbm_querycap(struct file *file, void *priv,
			   struct v4l2_capability *cap)
{
	strlcpy(cap->driver, MODULE_NAME, sizeof(cap->driver));
	strlcpy(cap->card, MODULE_NAME, sizeof(cap->card));
	snprintf(cap->bus_info, sizeof(cap->bus_info),
		 "platform:%s", MODULE_NAME);
	cap->device_caps = V4L2_CAP_VIDEO_M2M | V4L2_CAP_STREAMING;
	cap->capabilities = cap->device_caps | V4L2_CAP_DEVICE_CAPS;
	return 0;
};

static int hwsgbm_enum_fmt_output(struct file *file, void *priv,
				  struct v4l2_fmtdesc *f)
{
	if (f->index > ARRAY_SIZE(output_formats))
		return -EINVAL;
	*f = output_formats[f->index];
	return 0;
}

static int hwsgbm_enum_fmt_capture(struct file *file, void *priv,
				   struct v4l2_fmtdesc *f)
{
	if (f->index >= ARRAY_SIZE(capture_formats))
		return -EINVAL;
	*f = capture_formats[f->index];
	return 0;
}

static int hwsgbm_g_fmt_output(struct file *file, void *priv,
			       struct v4l2_format *f)
{
	struct hwsgbm_ctx *ctx = container_of(priv, struct hwsgbm_ctx, fh);

	f->fmt.pix.width = ctx->width;
	f->fmt.pix.height = ctx->height;
	f->fmt.pix.pixelformat = output_formats[ctx->outfmt].pixelformat;
	f->fmt.pix.field = V4L2_FIELD_ALTERNATE_LR;
	if (output_formats[ctx->outfmt].pixelformat == V4L2_PIX_FMT_GREY)
		f->fmt.pix.sizeimage = ctx->stride * ctx->height;
	else
		f->fmt.pix.sizeimage = ctx->stride * ctx->height * 3 / 2;

	f->fmt.pix.bytesperline = ctx->stride;
	f->fmt.pix.colorspace = V4L2_COLORSPACE_REC709;
	f->fmt.pix.flags = output_formats[ctx->outfmt].flags;

	return 0;
}

static int hwsgbm_g_fmt_capture(struct file *file, void *priv,
				struct v4l2_format *f)
{
	struct hwsgbm_ctx *ctx = container_of(priv, struct hwsgbm_ctx, fh);

	f->fmt.pix.width = ctx->disp_width;
	f->fmt.pix.height = ctx->disp_height;
	f->fmt.pix.pixelformat = capture_formats[ctx->capfmt].pixelformat;
	f->fmt.pix.field = V4L2_FIELD_NONE;
	f->fmt.pix.bytesperline = ctx->disp_stride;
	f->fmt.pix.sizeimage = ctx->disp_stride * round_up(ctx->disp_height,
							   BUF_DISP_VSIZE);
	f->fmt.pix.colorspace = V4L2_COLORSPACE_REC709;
	f->fmt.pix.flags = capture_formats[ctx->capfmt].flags;

	return 0;
}

static int hwsgbm_s_fmt_output(struct file *file, void *priv,
			       struct v4l2_format *f)
{
	struct hwsgbm_ctx *ctx = container_of(priv, struct hwsgbm_ctx, fh);
	unsigned int fmt = 0;

	if (ctx_is_active(ctx))
		return -EBUSY;
	if (f->fmt.pix.width % 2) {
		v4l2_printk(KERN_NOTICE, &ctx->drv_priv->v4l2_dev,
			    "Odd widths are not supported.\n");
		return -EINVAL;
	}
	ctx->width = f->fmt.pix.width;
	ctx->height = f->fmt.pix.height;
	while (fmt < ARRAY_SIZE(output_formats) &&
	       f->fmt.pix.pixelformat != output_formats[fmt].pixelformat)
		fmt++;

	if (fmt >= ARRAY_SIZE(output_formats))
		fmt = 0;

	ctx->outfmt = fmt;
	hwsgbm_calc_params(ctx);

	return hwsgbm_g_fmt_output(file, priv, f);
}

static int hwsgbm_s_fmt_capture(struct file *file, void *priv,
				struct v4l2_format *f)
{
	struct hwsgbm_ctx *ctx = container_of(priv, struct hwsgbm_ctx, fh);
	unsigned int fmt = 0;

	if (ctx_is_active(ctx))
		return -EBUSY;

	/* Ignore width & height. They are only set for output end. */

	while (fmt < ARRAY_SIZE(capture_formats) &&
	       f->fmt.pix.pixelformat != capture_formats[fmt].pixelformat)
		fmt++;

	if (fmt >= ARRAY_SIZE(capture_formats))
		fmt = 0;

	ctx->capfmt = fmt;

	return hwsgbm_g_fmt_capture(file, priv, f);
}

static const struct v4l2_ioctl_ops hwsgbm_ioctl_ops = {
	.vidioc_querycap = hwsgbm_querycap,

	.vidioc_enum_fmt_vid_out = hwsgbm_enum_fmt_output,
	.vidioc_g_fmt_vid_out    = hwsgbm_g_fmt_output,
	.vidioc_s_fmt_vid_out    = hwsgbm_s_fmt_output,

	.vidioc_enum_fmt_vid_cap = hwsgbm_enum_fmt_capture,
	.vidioc_g_fmt_vid_cap    = hwsgbm_g_fmt_capture,
	.vidioc_s_fmt_vid_cap    = hwsgbm_s_fmt_capture,

	.vidioc_reqbufs   = v4l2_m2m_ioctl_reqbufs,
	.vidioc_querybuf  = v4l2_m2m_ioctl_querybuf,
	.vidioc_qbuf      = v4l2_m2m_ioctl_qbuf,
	.vidioc_dqbuf     = v4l2_m2m_ioctl_dqbuf,
	.vidioc_expbuf    = v4l2_m2m_ioctl_expbuf,

	.vidioc_streamon  = v4l2_m2m_ioctl_streamon,
	.vidioc_streamoff = v4l2_m2m_ioctl_streamoff,

	.vidioc_subscribe_event = v4l2_ctrl_subscribe_event,
	.vidioc_unsubscribe_event = v4l2_event_unsubscribe
};

/*
 * Queue operations
 */

static int hwsgbm_queue_setup(struct vb2_queue *vq,
			      const struct v4l2_format *fmt,
			      unsigned int *nbuffers, unsigned int *nplanes,
			      unsigned int sizes[], void *alloc_ctxs[])
{
	struct hwsgbm_ctx *ctx = vb2_get_drv_priv(vq);

	switch (vq->type) {
	case V4L2_BUF_TYPE_VIDEO_OUTPUT:
		if (output_formats[ctx->outfmt].pixelformat
		    == V4L2_PIX_FMT_GREY)
			sizes[0] = ctx->stride * ctx->height;
		else
			sizes[0] = ctx->stride * ctx->height * 3 / 2;
		break;
	case V4L2_BUF_TYPE_VIDEO_CAPTURE:
		/* extra size so copying doesn't go out of bounds */
		sizes[0] = round_up(ctx->disp_height, BUF_DISP_VSIZE) *
			   ctx->disp_stride;
		break;
	default:
		WARN_ON(1);
		return -EINVAL;
	}
	*nplanes = 1;

	alloc_ctxs[0] = ctx->drv_priv->alloc_ctx;
	pr_devel("Request %d buffer(s) of size %d each.\n", *nbuffers,
		 sizes[0]);

	return 0;
}

static int hwsgbm_buf_prepare(struct vb2_buffer *vb)
{
	struct hwsgbm_ctx *ctx = vb2_get_drv_priv(vb->vb2_queue);
	unsigned int size;

	if (vb->vb2_queue->type == V4L2_BUF_TYPE_VIDEO_OUTPUT
	    && vb->v4l2_buf.field != V4L2_FIELD_LEFT
	    && vb->v4l2_buf.field != V4L2_FIELD_RIGHT)
		return -EINVAL;

	switch (vb->vb2_queue->type) {
	case V4L2_BUF_TYPE_VIDEO_OUTPUT:
		/*
		 * This function validates the buffer.
		 * While HWSGBM doesn't actually USE the chroma values in 4:2:0,
		 * and would work correctly if they were not allocated,
		 * the buffer not having enough space for chroma still means
		 * that the allocator screwed up, and we should still
		 * point that out.
		 */
		if (output_formats[ctx->outfmt].pixelformat
		    == V4L2_PIX_FMT_GREY)
			size = ctx->stride * ctx->height;
		else
			size = ctx->stride * ctx->height * 3 / 2;
		break;
	case V4L2_BUF_TYPE_VIDEO_CAPTURE:
		size = round_up(ctx->disp_height, BUF_DISP_VSIZE) *
		       ctx->disp_stride;
		break;
	default:
		WARN_ON(1);
		return -EINVAL;
	}

	if (vb2_plane_size(vb, 0) < size) {
		v4l2_printk(KERN_ALERT, &ctx->drv_priv->v4l2_dev,
			    "Data will not fit into plane (%lu < %u)\n",
			    vb2_plane_size(vb, 0), size);
		return -EINVAL;
	}

	return 0;
}

static void hwsgbm_buf_queue(struct vb2_buffer *vb)
{
	struct hwsgbm_ctx *ctx = vb2_get_drv_priv(vb->vb2_queue);

	v4l2_m2m_buf_queue(ctx->fh.m2m_ctx, vb);
}

static int hwsgbm_start_streaming(struct vb2_queue *vq, unsigned int count)
{
	int ret;
	struct hwsgbm_ctx *ctx = vb2_get_drv_priv(vq);
	if (ctx->nod > 127 && ctx->drv_priv->params.acc_direct != 1) {
		v4l2_printk(KERN_NOTICE, &ctx->drv_priv->v4l2_dev,
			    "Cannot use LR-check for NoD > 127\n");
		return -EINVAL;
	}
	if (vq->type == V4L2_BUF_TYPE_VIDEO_OUTPUT)
		return 0;
	if (ctx_activate(ctx) < 0)
		return -EBUSY;
	ret = hwsgbm_map_vram(ctx);
	if (ret) {
		ctx_deactivate(ctx);
		return ret;
	}
	ret = hwsgbm_vdma_request(ctx->drv_priv);
	if (ret)
		return ret;
	hwsgbm_writeconf(ctx);
	return 0;
}

static void hwsgbm_stop_streaming(struct vb2_queue *q)
{
	struct hwsgbm_ctx *ctx = vb2_get_drv_priv(q);
	struct vb2_buffer *vbuf;

	unsigned long flags;

	for (;;) {
		if (V4L2_TYPE_IS_OUTPUT(q->type))
			vbuf = v4l2_m2m_src_buf_remove(ctx->fh.m2m_ctx);
		else
			vbuf = v4l2_m2m_dst_buf_remove(ctx->fh.m2m_ctx);
		if (vbuf == NULL)
			break;
		spin_lock_irqsave(&ctx->drv_priv->irqlock, flags);
		v4l2_m2m_buf_done(vbuf, VB2_BUF_STATE_ERROR);
		spin_unlock_irqrestore(&ctx->drv_priv->irqlock, flags);
	}
	if (q->type == V4L2_BUF_TYPE_VIDEO_CAPTURE) {
		if (ctx_is_active(ctx)) {
			hwsgbm_unmap_vram(ctx);
			hwsgbm_vdma_release(ctx->drv_priv);
			ctx_deactivate(ctx);
		}
	}
}

static struct vb2_ops hwsgbm_qops = {
	.queue_setup     = hwsgbm_queue_setup,
	.buf_prepare     = hwsgbm_buf_prepare,
	.buf_queue       = hwsgbm_buf_queue,
	.start_streaming = hwsgbm_start_streaming,
	.stop_streaming  = hwsgbm_stop_streaming,
	.wait_prepare    = vb2_ops_wait_prepare,
	.wait_finish     = vb2_ops_wait_finish,
};

static int queue_init(void *priv, struct vb2_queue *src, struct vb2_queue *dst)
{
	struct hwsgbm_ctx *ctx = priv;
	int rc;

	src->type = V4L2_BUF_TYPE_VIDEO_OUTPUT;
	src->io_modes = VB2_MMAP | VB2_DMABUF;
	src->drv_priv = ctx;
	src->buf_struct_size = sizeof(struct v4l2_m2m_buffer);
	src->ops = &hwsgbm_qops;
	src->mem_ops = &vb2_dma_contig_memops;
	src->timestamp_flags = V4L2_BUF_FLAG_TIMESTAMP_COPY;
	src->lock = &ctx->drv_priv->mutex;

	rc = vb2_queue_init(src);
	if (rc)
		return rc;

	dst->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	dst->io_modes = VB2_MMAP | VB2_DMABUF;
	dst->drv_priv = ctx;
	dst->buf_struct_size = sizeof(struct v4l2_m2m_buffer);
	dst->ops = &hwsgbm_qops;
	dst->mem_ops = &vb2_dma_contig_memops;
	dst->timestamp_flags = V4L2_BUF_FLAG_TIMESTAMP_COPY;
	dst->lock = &ctx->drv_priv->mutex;

	return vb2_queue_init(dst);
}

static int hwsgbm_open(struct file *file)
{
	struct hwsgbm_priv *drv_priv = video_drvdata(file);
	struct hwsgbm_ctx *ctx = NULL;
	struct v4l2_ctrl *tmp_ctrl;
	int rc = 0;

	if (mutex_lock_interruptible(&drv_priv->mutex))
		return -ERESTARTSYS;

	ctx = kzalloc(sizeof(*ctx), GFP_KERNEL);
	if (!ctx) {
		rc = -ENOMEM;
		goto open_unlock;
	}

	ctx->drv_priv = drv_priv;
	ctx->state = HWSGBM_CTX_PASSIVE;

	ctx->width = 1280;
	ctx->height = 720;

	v4l2_fh_init(&ctx->fh, video_devdata(file));
	ctx->fh.ctrl_handler = &ctx->hdl;
	file->private_data = &ctx->fh;

	pr_devel("Initializing M2M context...\n");
	ctx->fh.m2m_ctx = v4l2_m2m_ctx_init(drv_priv->m2m_dev,
					    ctx, &queue_init);

	v4l2_ctrl_handler_init(&ctx->hdl, 1);
	tmp_ctrl = v4l2_ctrl_new_custom(&ctx->hdl, &ctrl_cfg[0],
					ctx);
	if (!tmp_ctrl) {
		v4l2_err(&drv_priv->v4l2_dev,
			 "Failed to create control: 0x%x\n", ctrl_cfg[0].id);
		rc = ctx->hdl.error;
		goto open_unlock;
	}
	v4l2_ctrl_add_handler(&ctx->hdl, &ctx->drv_priv->hdl, NULL);
	v4l2_ctrl_handler_setup(&ctx->hdl);

	if (IS_ERR(ctx->fh.m2m_ctx)) {
		rc = PTR_ERR(ctx->fh.m2m_ctx);
		pr_devel("Failed to initialize M2M context!\n");
		v4l2_ctrl_handler_free(&ctx->hdl);
		kfree(ctx);
		goto open_unlock;
	}

	v4l2_fh_add(&ctx->fh);

open_unlock:
	mutex_unlock(&drv_priv->mutex);
	return rc;
}

static int hwsgbm_release(struct file *file)
{
	struct hwsgbm_ctx *ctx
		= container_of(file->private_data, struct hwsgbm_ctx, fh);

	hwsgbm_stop_streaming(v4l2_m2m_get_vq(ctx->fh.m2m_ctx,
					      V4L2_BUF_TYPE_VIDEO_CAPTURE));
	hwsgbm_stop_streaming(v4l2_m2m_get_vq(ctx->fh.m2m_ctx,
					      V4L2_BUF_TYPE_VIDEO_OUTPUT));

	v4l2_fh_del(&ctx->fh);
	v4l2_fh_exit(&ctx->fh);
	v4l2_ctrl_handler_free(&ctx->hdl);
	mutex_lock(&ctx->drv_priv->mutex);
	v4l2_m2m_ctx_release(ctx->fh.m2m_ctx);
	mutex_unlock(&ctx->drv_priv->mutex);
	kfree(ctx);
	return 0;
}

static const struct v4l2_file_operations hwsgbm_fops = {
	.owner   = THIS_MODULE,
	.open    = hwsgbm_open,
	.release = hwsgbm_release,
	.poll    = v4l2_m2m_fop_poll,
	.unlocked_ioctl = video_ioctl2,
	.mmap    = v4l2_m2m_fop_mmap
};

static struct video_device const hwsgbm_video_device = {
	.name      = MODULE_NAME,
	.vfl_dir   = VFL_DIR_M2M,
	.fops      = &hwsgbm_fops,
	.ioctl_ops = &hwsgbm_ioctl_ops,
	.minor     = -1,
	.release = video_device_release_empty
};

static int hwsgbm_probe(struct platform_device *pdev)
{
	struct hwsgbm_priv *drv_priv;
	struct resource *devres = NULL;
	int ret, irq;

	drv_priv = devm_kzalloc(&pdev->dev, sizeof(*drv_priv), GFP_KERNEL);
	if (!drv_priv)
		return -ENOMEM;

	ret = v4l2_device_register(&pdev->dev, &drv_priv->v4l2_dev);
	if (ret)
		return ret;

	spin_lock_init(&drv_priv->irqlock);
	mutex_init(&drv_priv->mutex);

	devres = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!devres) {
		ret = -EINVAL;
		dev_err(&pdev->dev, "Failed to find memory resource\n");
		goto unreg_v4l2;
	}

	drv_priv->regs = devm_ioremap_resource(&pdev->dev, devres);
	if (IS_ERR(drv_priv->regs)) {
		ret = PTR_ERR(drv_priv->regs);
		dev_err(&pdev->dev, "Failed to remap mem resource\n");
		goto unreg_v4l2;
	}

	drv_priv->vrampool = of_get_named_gen_pool(pdev->dev.of_node,
						   "vram", 0);
	if (drv_priv->vrampool == NULL) {
		ret = -ENOENT;
		dev_err(&pdev->dev, "Failed to get VRAM resource\n");
		goto unreg_v4l2;
	}

	drv_priv->vram_start = hwsgbm_get_vram_base(pdev);
	if (drv_priv->vram_start == 0) {
		ret = -ENOENT;
		goto unreg_v4l2;
	}

	irq = platform_get_irq(pdev, 0);
	if (irq < 0) {
		dev_err(&pdev->dev, "Cannot parse IRQ from platform\n");
		goto unreg_v4l2;
	}
	ret = devm_request_irq(&pdev->dev, irq, hwsgbm_irq_handler,
			       IRQF_SHARED, "hwsgbm", drv_priv);
	if (ret) {
		dev_err(&pdev->dev, "Failed to map IRQ to handler\n");
		goto unreg_v4l2;
	}

	drv_priv->vfd = hwsgbm_video_device;
	drv_priv->vfd.lock = &drv_priv->mutex;
	drv_priv->vfd.v4l2_dev = &drv_priv->v4l2_dev;
	ret = video_register_device(&drv_priv->vfd, VFL_TYPE_GRABBER, -1);
	if (ret) {
		v4l2_err(&drv_priv->v4l2_dev,
			 "Failed to register video device\n");
		goto unreg_v4l2;
	}

	drv_priv->alloc_ctx = vb2_dma_contig_init_ctx(&pdev->dev);
	if (IS_ERR(drv_priv->alloc_ctx)) {
		v4l2_err(&drv_priv->v4l2_dev, "Failed to alloc vb2 context\n");
		goto unreg_vfd;
	}

	video_set_drvdata(&drv_priv->vfd, drv_priv);
	v4l2_info(&drv_priv->v4l2_dev, "Device hwsgbm registered as /dev/video%d\n",
		drv_priv->vfd.num);
	platform_set_drvdata(pdev, drv_priv);
	drv_priv->m2m_dev = v4l2_m2m_init(&hwsgbm_m2m_ops);
	if (IS_ERR(drv_priv->m2m_dev)) {
		v4l2_err(&drv_priv->v4l2_dev,
			 "Failed to init mem2mem device\n");
		ret = PTR_ERR(drv_priv->m2m_dev);
		goto unreg_dmactx;
	}

	ret = hwsgbm_ctrl_init(drv_priv);
	if (ret) {
		v4l2_err(&drv_priv->v4l2_dev, "Failed to init ctrl handler\n");
		goto unreg_m2mdev;
	}

	return 0;
unreg_m2mdev:
	v4l2_m2m_release(drv_priv->m2m_dev);
unreg_dmactx:
	vb2_dma_contig_cleanup_ctx(drv_priv->alloc_ctx);
unreg_vfd:
	video_unregister_device(&drv_priv->vfd);
unreg_v4l2:
	v4l2_device_unregister(&drv_priv->v4l2_dev);

	return ret;
}

static int hwsgbm_remove(struct platform_device *pdev)
{
	struct hwsgbm_priv *drv_priv =
		(struct hwsgbm_priv *)platform_get_drvdata(pdev);

	v4l2_ctrl_handler_free(&drv_priv->hdl);
	v4l2_info(&drv_priv->v4l2_dev, "Removing " MODULE_NAME);
	v4l2_m2m_release(drv_priv->m2m_dev);
	vb2_dma_contig_cleanup_ctx(drv_priv->alloc_ctx);
	video_unregister_device(&drv_priv->vfd);
	v4l2_device_unregister(&drv_priv->v4l2_dev);

	return 0;
}

static struct platform_device_id hwsgbm_platform_ids[] = {
	{ .name = "hwsgbm", .driver_data = 0 },
	{ /* sentinel */ }
};
MODULE_DEVICE_TABLE(platform, hwsgbm_platform_ids);

#ifdef CONFIG_OF
static const struct of_device_id hwsgbm_dt_ids[] = {
	{ .compatible = "elvees,hwsgbm", .data = NULL },
	{ /* sentinel */ }
};
MODULE_DEVICE_TABLE(of, hwsgbm_dt_ids);
#endif

static struct platform_driver hwsgbm_driver = {
	.probe = hwsgbm_probe,
	.remove = hwsgbm_remove,
	.driver = {
		.name = MODULE_NAME,
		.owner = THIS_MODULE,
		.of_match_table = of_match_ptr(hwsgbm_dt_ids)
	},
	.id_table = hwsgbm_platform_ids
};

module_platform_driver(hwsgbm_driver);
