/*
 * HWSGBM driver - private header
 *
 * Copyright 2016 ELVEES NeoTek JSC
 * Author: Oleg Kitain <okitain@elvees.com>
 */

#ifndef HWSGBM_DRIVER_H
#define HWSGBM_DRIVER_H

#include <linux/genalloc.h>
#include <linux/spinlock.h>
#include <linux/mutex.h>

#include <media/v4l2-mem2mem.h>
#include <media/v4l2-device.h>
#include <media/v4l2-ctrls.h>

#define SGBM_CONTROL		0x00
#define SGBM_STATUS		0x04
#define SGBM_IRQMASK		0x08
#define SGBM_IRQTEST		0x0C
#define SGBM_SL0		BIT(0)
#define SGBM_SL1		BIT(1)
#define SGBM_SR0		BIT(2)
#define SGBM_SR1		BIT(3)
#define SGBM_R0			BIT(4)
#define SGBM_R1			BIT(5)
#define SGBM_ALL_BUFS		GENMASK(0, 5)
#define SGBM_ERROR		BIT(24)
#define SGBM_EFRAME		BIT(25)

#define SGBM_EVTMAP		0x10
#define SGBM_EVTMASK		0x14
#define EVT_SHIFT(n)		(5 * (n))

#define SGBM_CONFIG		0x30
#define ACC_DIRECT_MASK		GENMASK(1, 0)
#define NOD_MASK		GENMASK(15, 8)
#define NOD_SHIFT		8
#define MAX_LR_DIFF_MASK	GENMASK(23, 16)
#define MAX_LR_DIFF_SHIFT	16
#define PEAK_EN			BIT(24)
#define INTERP_EN		BIT(25)
#define MEDIAN_EN		BIT(26)
#define SGBM_EN			BIT(31)

#define SGBM_DMAP_PARAM		0x34
#define DMAP_HSIZE_MASK		GENMASK(10, 0)
#define DMAP_VSIZE_MASK		GENMASK(26, 16)
#define DMAP_VSIZE_SHIFT	16

#define SGBM_FILT_PARAM		0x38
#define SGBM_PENALTY		0x3C
#define PENALTY1_MASK		GENMASK(10, 0)
#define PENALTY2_MASK		GENMASK(26, 16)
#define PENALTY2_SHIFT		16

#define SGBM_PEAK_PARAM		0x40
#define MAXPEAK_MASK		GENMASK(5, 0)

#define SGBM_INTERP_PARAM	0x44
#define INTERP_RADIUS_MASK	GENMASK(3, 0)
#define EDGE_EN			BIT(8)
#define ISLAND_EN		BIT(9)
#define NUM_VALID_MASK		GENMASK(15, 12)
#define NUM_VALID_SHIFT		12

#define SGBM_MEDIAN_PARAM	0x50
#define STAGES_NUM_MASK		GENMASK(3, 0)

#define SGBM_MEDIAN_REP1	0x54
#define SGBM_MEDIAN_REP2	0x58
#define SGBM_LSRC0_BASE		0x60
#define SGBM_LSRC1_BASE		0x64
#define SGBM_RSRC0_BASE		0x68
#define SGBM_RSRC1_BASE		0x6C
#define SGBM_SRC_PARAM		0x70
#define SGBM_LDRAW0_BASE	0x80
#define SGBM_LDRAW1_BASE	0x84
#define SGBM_RDRAW_BASE		0x88
#define SGBM_DRAW_PARAM		0x8C
#define SGBM_DPRE0_BASE		0xA0
#define SGBM_DPRE1_BASE		0xA4
#define SGBM_DPRE_PARAM		0xA8
#define SGBM_DDEN0_BASE		0xB0
#define SGBM_DDEN1_BASE		0xB4
#define SGBM_DDEN_PARAM		0xB8
#define SGBM_ACML_BASE		0xC0
#define SGBM_ACMR_BASE		0xC4
#define SGBM_LTMP_BASE		0xD0
#define SGBM_RTMP_BASE		0xD4
#define SGBM_TMP_PARAM		0xD8

#define SGBM_INTUSE_MEDIAN_R1	0xF8
#define SGBM_INTUSE_MEDIAN_R2	0xFC

#define BUF_SRC_VSIZE	4
#define BUF_DISP_VSIZE	8
#define VDMA_CHAN_NUM	6

struct hwsgbm_priv {
	struct v4l2_device v4l2_dev;
	struct video_device vfd;

	void __iomem *regs;
	void *vram_virt;
	phys_addr_t vram_phys;
	phys_addr_t vram_start;
	size_t vram_len;
	int32_t srcl0_ofst;
	int32_t srcr0_ofst;
	int32_t srcr1_ofst;
	int32_t srcl1_ofst;
	int32_t disp0_ofst;
	int32_t disp1_ofst;

	int vdma_chmap[VDMA_CHAN_NUM];

	spinlock_t irqlock;
	struct mutex mutex;
	struct gen_pool *vrampool;
	struct v4l2_m2m_dev *m2m_dev;
	struct vb2_alloc_ctx *alloc_ctx;

	struct v4l2_ctrl_handler hdl;
	/* Abort requested by m2m */
	int aborting;

	struct {
		uint8_t acc_direct;
		uint8_t peak_en;
		uint8_t interpolate_en;
		uint8_t median_en;
	} params;

	atomic_t active;
};

enum hwsgbm_ctx_state {
	HWSGBM_CTX_PASSIVE = 0,
	HWSGBM_CTX_ACTIVE = 1,
};

struct hwsgbm_ctx {
	struct hwsgbm_priv *drv_priv;
	struct v4l2_ctrl_handler hdl;
	struct v4l2_fh fh;
	uint32_t width, height, disp_width, disp_height;
	uint32_t stride, disp_stride;
	uint32_t capfmt, outfmt;
	uint8_t nod;
	enum hwsgbm_ctx_state state;
};

static inline u32 hwsgbm_read(struct hwsgbm_priv const *const drv_priv,
			      off_t const reg)
{
	return ioread32(drv_priv->regs + reg);
};

static inline void hwsgbm_write(u32 const value,
				struct hwsgbm_priv const *const drv_priv,
				off_t const reg)
{
	iowrite32(value, drv_priv->regs + reg);
};

#endif /* HWSGBM_DRIVER_H */
