/*
 * HWSGBM driver - VDMA programming functions
 *
 * Copyright 2016 ELVEES NeoTek JSC
 * Author: Oleg Kitain <okitain@elvees.com>
 */

#include <linux/vdma.h>

#include "hwsgbm-driver.h"
#include "hwsgbm-vdma.h"

int hwsgbm_vdma_request(struct hwsgbm_priv *drv_priv)
{
	int i, j;

	for (i = 0; i < VDMA_CHAN_NUM; i++) {
		drv_priv->vdma_chmap[i] = vdma_request_channel();
		if (drv_priv->vdma_chmap[i] < 0) {
			for (j = i - 1; j >= 0; j--)
				vdma_release_channel(drv_priv->vdma_chmap[j]);
			return drv_priv->vdma_chmap[i];
		}
	}
	return 0;
}

void hwsgbm_vdma_release(struct hwsgbm_priv *drv_priv)
{
	int i;

	for (i = 0; i < VDMA_CHAN_NUM; i++)
		vdma_release_channel(drv_priv->vdma_chmap[i]);
}

/* TODO: rework for variable stride */
static void vdma_chan_conf(struct hwsgbm_priv *drv_priv,
			   dma_addr_t emem_addr,
			   uint32_t imem_addr,
			   uint32_t bufsize,
			   uint8_t sgbm_chnum,
			   uint8_t direction,
			   uint32_t buf_totnum)
{
	uint32_t cfg_reg, cnt;
	int vdma_ch = drv_priv->vdma_chmap[sgbm_chnum];

	vdma_write(vdma_ch, VDMA_CHAN_RUN, 0);
	vdma_write(vdma_ch, VDMA_CHAN_DONE, 0);
	vdma_write(vdma_ch, VDMA_CHAN_IMRDY, 0);
	vdma_write(vdma_ch, VDMA_CHAN_AECUR, emem_addr);
	vdma_write(vdma_ch, VDMA_CHAN_A0E, emem_addr);
	vdma_write(vdma_ch, VDMA_CHAN_AICUR, imem_addr);
	vdma_write(vdma_ch, VDMA_CHAN_A0I, imem_addr);
	vdma_write(vdma_ch, VDMA_CHAN_BEIDX, 0);
	vdma_write(vdma_ch, VDMA_CHAN_CEIDX, 0);
	/* skip over other buffer */
	vdma_write(vdma_ch, VDMA_CHAN_HEIDX, bufsize);
	vdma_write(vdma_ch, VDMA_CHAN_VEIDX, 0);
	vdma_write(vdma_ch, VDMA_CHAN_BIIDX, 0);
	vdma_write(vdma_ch, VDMA_CHAN_CIIDX, 0);
	/* VRAM: write over the same location */
	vdma_write(vdma_ch, VDMA_CHAN_HIIDX, -bufsize);

	vdma_write(vdma_ch, VDMA_CHAN_VIIDX, 0);

	cnt = FIELD_PREP(VDMA_ACNT_ACNT, bufsize / 16 - 1) |
	      FIELD_PREP(VDMA_ACNT_ARLD, bufsize / 16 - 1);
	vdma_write(vdma_ch, VDMA_CHAN_ACNT, cnt);

	vdma_write(vdma_ch, VDMA_CHAN_BCCNT, 0);

	cnt = FIELD_PREP(VDMA_HVXCNT_HCNT, buf_totnum - 1) |
	      FIELD_PREP(VDMA_HVXCNT_HRLD, buf_totnum - 1);
	vdma_write(vdma_ch, VDMA_CHAN_HVECNT, cnt);
	vdma_write(vdma_ch, VDMA_CHAN_HVICNT, cnt);

	vdma_write(vdma_ch, VDMA_CHAN_RUN, 1);
	vdma_write(vdma_ch, VDMA_CHAN_DONE, 0);
	vdma_write(vdma_ch, VDMA_CHAN_IMRDY, 0);
	cfg_reg = (direction ? VDMA_CFG_DIR : 0)
		| VDMA_CFG_CYCLE
		| VDMA_CFG_BRST_AE
		| VDMA_CFG_BRST_AI
		| VDMA_CFG_BC_DONE
		| VDMA_CFG_DONE_MSK
		| FIELD_PREP(VDMA_CFG_SIZE, 4);
	vdma_write(vdma_ch, VDMA_CHAN_CFG, cfg_reg);
}

void hwsgbm_conf_vdma(struct hwsgbm_ctx *ctx,
		      dma_addr_t srcl_phys,
		      dma_addr_t srcr_phys,
		      dma_addr_t dst_phys)
{
	struct hwsgbm_priv *drv_priv = ctx->drv_priv;
	uint32_t count_src0, count_src1, count_dst0, count_dst1;

	count_src0 = DIV_ROUND_UP(ctx->height, BUF_SRC_VSIZE * 2);
	count_src1 = DIV_ROUND_UP(ctx->height, BUF_SRC_VSIZE) / 2;

	count_dst0 = DIV_ROUND_UP(ctx->disp_height,
				  BUF_DISP_VSIZE * 2);
	count_dst1 = DIV_ROUND_UP(ctx->disp_height,
				  BUF_DISP_VSIZE) / 2;

	vdma_chan_conf(drv_priv, srcl_phys,
		       drv_priv->srcl0_ofst + drv_priv->vram_phys,
		       ctx->stride * BUF_SRC_VSIZE, 0, 0,
		       count_src0);
	vdma_chan_conf(drv_priv,
		       srcl_phys + ctx->stride * BUF_SRC_VSIZE,
		       drv_priv->srcl1_ofst + drv_priv->vram_phys,
		       ctx->stride * BUF_SRC_VSIZE, 1, 0,
		       count_src1);
	vdma_chan_conf(drv_priv, srcr_phys,
		       drv_priv->srcr0_ofst + drv_priv->vram_phys,
		       ctx->stride * BUF_SRC_VSIZE, 2, 0,
		       count_src0);
	vdma_chan_conf(drv_priv,
		       srcr_phys + ctx->stride * BUF_SRC_VSIZE,
		       drv_priv->srcr1_ofst + drv_priv->vram_phys,
		       ctx->stride * BUF_SRC_VSIZE, 3, 0,
		       count_src1);

	vdma_chan_conf(drv_priv, dst_phys,
		       drv_priv->disp0_ofst + drv_priv->vram_phys,
		       ctx->disp_stride * BUF_DISP_VSIZE, 4, 1,
		       count_dst0);
	vdma_chan_conf(drv_priv,
		       dst_phys + ctx->disp_stride * BUF_DISP_VSIZE,
		       drv_priv->disp1_ofst + drv_priv->vram_phys,
		       ctx->disp_stride * BUF_DISP_VSIZE, 5, 1,
		       count_dst1);
}

void hwsgbm_vdma_run(struct hwsgbm_priv *drv_priv)
{
	uint32_t reg_irqmask;

	reg_irqmask = hwsgbm_read(drv_priv, SGBM_IRQMASK);
	hwsgbm_write(reg_irqmask & ~SGBM_EFRAME, drv_priv, SGBM_IRQMASK);

	vdma_write(drv_priv->vdma_chmap[4], VDMA_CHAN_DONE, 1);
	vdma_write(drv_priv->vdma_chmap[5], VDMA_CHAN_DONE, 1);
	vdma_write(drv_priv->vdma_chmap[0], VDMA_CHAN_DONE, 0);
	vdma_write(drv_priv->vdma_chmap[1], VDMA_CHAN_DONE, 0);
	vdma_write(drv_priv->vdma_chmap[2], VDMA_CHAN_DONE, 0);
	vdma_write(drv_priv->vdma_chmap[3], VDMA_CHAN_DONE, 0);

	vdma_write(drv_priv->vdma_chmap[0], VDMA_CHAN_IMRDY, 1);
	vdma_write(drv_priv->vdma_chmap[2], VDMA_CHAN_IMRDY, 1);
	vdma_write(drv_priv->vdma_chmap[1], VDMA_CHAN_IMRDY, 1);
	vdma_write(drv_priv->vdma_chmap[3], VDMA_CHAN_IMRDY, 1);
}

/*
 * This function exists because of
 * https://callisto.elvees.com/redmine/issues/2010
 */
void hwsgbm_vdma_done(struct hwsgbm_priv *drv_priv)
{
	uint32_t r1_st, r2_st;
	int i;

	/* Since this is called from IRQ, function can't sleep */
	while (1) {
		r1_st = vdma_read(drv_priv->vdma_chmap[4], VDMA_CHAN_DONE);
		r2_st = vdma_read(drv_priv->vdma_chmap[5], VDMA_CHAN_DONE);
		if (r1_st && r2_st)
			break;
	}
	for (i = 0; i < VDMA_CHAN_NUM; i++)
		vdma_write(drv_priv->vdma_chmap[i], VDMA_CHAN_RUN, 0);
}
