/*
 * HWSGBM driver - VRAM allocation functions
 *
 * Copyright 2016 ELVEES NeoTek JSC
 * Author: Oleg Kitain <okitain@elvees.com>
 */

#include <linux/genalloc.h>
#include <linux/list.h>
#include <linux/of_address.h>
#include <linux/platform_device.h>
#include <linux/rculist.h>
#include <linux/slab.h>
#include <linux/sizes.h>

#include "hwsgbm-driver.h"
#include "hwsgbm-vram.h"

#define BANK_SIZE SZ_64K

phys_addr_t hwsgbm_get_vram_base(struct platform_device *pdev)
{
	int ret;
	struct device_node *vram_np;
	struct resource res;

	vram_np = of_parse_phandle(pdev->dev.of_node, "vram", 0);
	if (!vram_np) {
		dev_err(&pdev->dev, "Failed to dereference VRAM phandle\n");
		return 0;
	}
	ret = of_address_to_resource(vram_np, 0, &res);
	if (ret) {
		dev_err(&pdev->dev, "Failed to get VRAM address\n");
		return 0;
	}
	return res.start;
}

struct space_chunk {
	struct list_head list;
	uint32_t start; /* offset from beginning of allocated vram */
	size_t len;
	/* Members below are not used for freespace chunks */
	off_t reg; /* Register to write to, 0 if unneeded */
	int32_t *priv_ptr; /* Member in priv to write to, NULL if unneeded */
	int valid; /* 1 if no errors encountered, 0 if some are. */
};

static inline int32_t vram_iaddr(int32_t offset, int32_t vram_offset)
{
	int32_t bybank_iaddr = vram_offset + offset;

	return ((bybank_iaddr & 0x1F0000) >> 16)
		| ((bybank_iaddr & 0xFFF0) << 1);
}

static struct space_chunk *find_valid_freespace(struct list_head *freespace,
						uint32_t start,
						size_t len)
{
	struct space_chunk *ptr;

	if (len == 0)
		return NULL;

	list_for_each_entry(ptr, freespace, list) {
		if (start >= ptr->start &&
		    (start < ptr->start + ptr->len)) {
			if (start + len <= ptr->start + ptr->len)
				return ptr;
			else
				return NULL;
		}
	}
	return NULL;
}

/**
 * claim_space - claim some VRAM space as used
 *
 * start - start of claimed space
 * len - size of claimed space
 *
 * Claiming means space will no longer feature in the list of freespace chunks
 */
static int claim_space(struct list_head *freespace,
		       struct list_head *filled_space,
		       uint32_t start,
		       size_t len,
		       off_t reg,
		       int32_t *priv_ptr)
{
	struct space_chunk *ptr;
	struct space_chunk *new_filled;
	int old_valid;

	/* filled_space is not validated for cross-allocation or sorted.
	 * It's used for logging claim_space requests.
	 */
	new_filled = kmalloc(sizeof(struct space_chunk), GFP_KERNEL);
	new_filled->start = start;
	new_filled->len = len;
	new_filled->reg = reg;
	new_filled->priv_ptr = priv_ptr;
	/* We don't actually want to check output of every claim_space.
	 * So if a claim is invalid, every next claim is also invalid,
	 * and therefore ignored.
	 */
	new_filled->valid = 1;
	if (list_empty(filled_space))
		old_valid = 1;
	else {
		ptr = list_last_entry(filled_space, struct space_chunk, list);
		old_valid = ptr->valid;
	}
	list_add_tail(&new_filled->list, filled_space);

	ptr = find_valid_freespace(freespace, start, len);
	if (ptr == NULL || !old_valid) {
		new_filled->valid = 0;
		return -EINVAL;
	}

	/* There are three cases that can happen when you are claiming space:
	 * Case 1. Available space is shrunk: start or len are modified.
	 * Case 2. Available space is split into two: additional chunk created.
	 * Case 3. Available space exactly claimed, and therefore will be gone.
	 */
	if (ptr->start == start || ptr->start + ptr->len == start + len) {
		if (ptr->start == start && ptr->len == len) {
			/* Case 3. */
			list_del(&ptr->list);
			kfree(ptr);
		} else {
			/* Case 1. */
			if (ptr->start == start) {
				ptr->len = ptr->len - len;
				ptr->start = start + len;
			} else {
				ptr->len = start - ptr->start;
			}
		}
	} else {
		/* Case 2. */
		uint32_t fin;
		struct space_chunk *newspace;

		fin = ptr->start + ptr->len;
		ptr->len = start - ptr->start;

		newspace = kmalloc(sizeof(struct space_chunk), GFP_KERNEL);
		list_add(&newspace->list, &ptr->list);

		newspace->start = start + len;
		newspace->len = fin - newspace->start;
	}

	return 0;
};

static int32_t autoclaim_space(struct list_head *freespace,
			       struct list_head *filled_space,
			       size_t len, off_t reg, int32_t *priv_ptr)
{
	struct space_chunk *ptr;
	int ret;
	int32_t start;

	list_for_each_entry(ptr, freespace, list) {
		if (len <= ptr->len) {
			start = ptr->start;
			ret = claim_space(freespace, filled_space,
					  ptr->start, len,
					  reg, priv_ptr);
			if (ret)
				return ret;
			return start;
		}
	}
	return -ENOMEM;
};

static size_t get_claimed_total(struct list_head *freespace)
{
	struct space_chunk *ptr;

	ptr = list_last_entry(freespace, struct space_chunk, list);
	return ptr->start;
};

static int check_chain_valid(struct list_head *filled_space)
{
	struct space_chunk *ptr;

	ptr = list_last_entry(filled_space, struct space_chunk, list);
	return ptr->valid;
}

static int write_all_filled(struct hwsgbm_priv *drv_priv,
			    struct list_head *filled_space)
{
	struct space_chunk *ptr;
	uint32_t alloc_offset;

	if (!check_chain_valid(filled_space))
		return -EINVAL;
	drv_priv->vram_virt = gen_pool_dma_alloc(
		drv_priv->vrampool,
		drv_priv->vram_len,
		&drv_priv->vram_phys);
	if (!drv_priv->vram_virt)
		return -ENOMEM;
	alloc_offset = drv_priv->vram_phys - drv_priv->vram_start;

	list_for_each_entry(ptr, filled_space, list) {
		if (!ptr->valid)
			return -EINVAL;
		if (ptr->reg)
			hwsgbm_write(vram_iaddr(ptr->start, alloc_offset),
				     drv_priv, ptr->reg);
		if (ptr->priv_ptr != NULL)
			*ptr->priv_ptr = ptr->start;
	}
	return 0;
}

int hwsgbm_map_vram(struct hwsgbm_ctx *ctx)
{
	struct space_chunk *tmp, *store;
	struct hwsgbm_priv *drv_priv = ctx->drv_priv;
	unsigned int i, num_lbank, num_rbank, acm_mult, disp_hsize, postproc;
	unsigned int acm_toadd;
	int ret;
	uint32_t acml_offset, acmr_offset;
	size_t acm_size = 0, tmp_size = 0, src_size = 0,
		dispraw_size = 0, disp_pr_size = 0, disp_dense_size = 0;

	LIST_HEAD(freespace);
	LIST_HEAD(filled_space);

	tmp = kmalloc(sizeof(struct space_chunk), GFP_KERNEL);
	list_add(&tmp->list, &freespace);

	tmp->start = 0;
	tmp->len = UINT_MAX;

	acm_mult = ctx->nod < 64 ? 1 : 2;
	acm_toadd = (DIV_ROUND_UP(ctx->width - 2, 16)
		- DIV_ROUND_UP(ctx->disp_width, 16) > ctx->nod / 16 + 1) ?
		16 : 0;
	acm_size = (round_up(ctx->disp_width, 16) + acm_toadd) *
		acm_mult * 16;
	if (ctx->nod == 15) {
		num_lbank = 2;
		num_rbank = 2;
	} else if (ctx->nod == 31) {
		num_lbank = 4;
		num_rbank = 4;
	} else if (ctx->nod == 47) {
		num_lbank = 5;
		num_rbank = 5;
	} else if (ctx->nod <= 127) {
		num_lbank = 7;
		num_rbank = 7;
	} else if (ctx->nod == 143) {
		num_lbank = 7;
		num_rbank = 2;
	} else if (ctx->nod == 159) {
		num_lbank = 7;
		num_rbank = 4;
	} else if (ctx->nod == 175) {
		num_lbank = 7;
		num_rbank = 5;
	} else {
		num_lbank = 7;
		num_rbank = 7;
	}
	if (drv_priv->params.acc_direct == 1 && ctx->nod <= 127)
		num_rbank = 0;
	if (drv_priv->params.acc_direct == 0) {
		num_rbank = 0;
		num_lbank = 0;
	}
	/* Claiming space for ACML.
	 * Since we are operating on a virtual chunk we allocate later,
	 * we can just put the offset at 0
	 */
	acml_offset = 0;
	for (i = 0; i < num_lbank; i++) {
		claim_space(&freespace, &filled_space,
			    acml_offset + i * BANK_SIZE,
			    acm_size,
			    i ? 0 : SGBM_ACML_BASE,
			    NULL);
	}

	/* for ACMR */
	if (acm_size <= BANK_SIZE / 2) {
		acmr_offset = acml_offset + acm_size;
	} else {
		acmr_offset = acml_offset + num_lbank * BANK_SIZE;
		/* packing ACMR RIGHT AFTER ACML can lead to mem bottleneck */
	}
	for (i = 0; i < num_rbank; i++) {
		claim_space(&freespace, &filled_space,
			    acmr_offset + i * BANK_SIZE,
			    acm_size,
			    i ? 0 : SGBM_ACMR_BASE,
			    NULL);
	}

	/* For TMP buffers */
	tmp_size = round_up(ctx->width - 2, 16) * 8;
	hwsgbm_write(8, drv_priv, SGBM_TMP_PARAM);
	autoclaim_space(&freespace, &filled_space,
			tmp_size,
			SGBM_LTMP_BASE, NULL);
	autoclaim_space(&freespace, &filled_space,
			tmp_size,
			SGBM_RTMP_BASE, NULL);

	/* For SRC buffers */
	src_size = round_up(ctx->width, 16) * BUF_SRC_VSIZE;
	hwsgbm_write(BUF_SRC_VSIZE, drv_priv, SGBM_SRC_PARAM);
	autoclaim_space(&freespace, &filled_space,
			src_size,
			SGBM_LSRC0_BASE,
			&drv_priv->srcl0_ofst);
	autoclaim_space(&freespace, &filled_space,
			src_size,
			SGBM_RSRC0_BASE,
			&drv_priv->srcr0_ofst);
	autoclaim_space(&freespace, &filled_space,
			src_size,
			SGBM_LSRC1_BASE,
			&drv_priv->srcl1_ofst);
	autoclaim_space(&freespace, &filled_space,
			src_size,
			SGBM_RSRC1_BASE,
			&drv_priv->srcr1_ofst);
	disp_hsize = round_up(ctx->disp_width, 16);
	postproc = drv_priv->params.peak_en || drv_priv->params.median_en
		|| drv_priv->params.interpolate_en;
	/* For DISP_RAW buffers */
	if (drv_priv->params.acc_direct == 1 && postproc) {
		dispraw_size = disp_hsize * 48;
		hwsgbm_write(48, drv_priv, SGBM_DRAW_PARAM);
	} else {
		dispraw_size = disp_hsize * BUF_DISP_VSIZE;
		if (drv_priv->params.acc_direct == 1)
			hwsgbm_write(BUF_DISP_VSIZE * 2, drv_priv,
				     SGBM_DRAW_PARAM);
		else
			hwsgbm_write(BUF_DISP_VSIZE, drv_priv,
				     SGBM_DRAW_PARAM);
	}

	if (drv_priv->params.acc_direct == 1 && !postproc) {
		autoclaim_space(&freespace, &filled_space,
				dispraw_size,
				SGBM_LDRAW0_BASE,
				&drv_priv->disp0_ofst);
		autoclaim_space(&freespace, &filled_space,
				dispraw_size,
				SGBM_LDRAW1_BASE,
				&drv_priv->disp1_ofst);
	} else {
		autoclaim_space(&freespace, &filled_space,
				dispraw_size,
				SGBM_LDRAW0_BASE,
				NULL);
		if (drv_priv->params.acc_direct != 1) {
			autoclaim_space(&freespace, &filled_space,
					dispraw_size,
					SGBM_RDRAW_BASE,
					NULL);
		}
	}
	/* For DISP_PRECISE buffers */
	if (drv_priv->params.acc_direct != 1) {
		if (postproc) {
			disp_pr_size = disp_hsize * 48;
			hwsgbm_write(48, drv_priv, SGBM_DPRE_PARAM);
			autoclaim_space(&freespace, &filled_space,
					disp_pr_size,
					SGBM_DPRE0_BASE,
					NULL);
		} else {
			disp_pr_size = disp_hsize * BUF_DISP_VSIZE;
			hwsgbm_write(BUF_DISP_VSIZE * 2,
				     drv_priv, SGBM_DPRE_PARAM);
			autoclaim_space(&freespace, &filled_space,
					disp_pr_size,
					SGBM_DPRE0_BASE,
					&drv_priv->disp0_ofst);
			autoclaim_space(&freespace, &filled_space,
					disp_pr_size,
					SGBM_DPRE1_BASE,
					&drv_priv->disp1_ofst);
		}
	}
	/*For DISP_DENSE buffers */
	if (postproc) {
		disp_dense_size = disp_hsize * BUF_DISP_VSIZE;
		hwsgbm_write(BUF_DISP_VSIZE, drv_priv, SGBM_DDEN_PARAM);
		autoclaim_space(&freespace, &filled_space,
				disp_dense_size,
				SGBM_DDEN0_BASE,
				&drv_priv->disp0_ofst);
		autoclaim_space(&freespace, &filled_space,
				disp_dense_size,
				SGBM_DDEN1_BASE,
				&drv_priv->disp1_ofst);
	}
	drv_priv->vram_len = get_claimed_total(&freespace);

	if (!num_rbank)
		hwsgbm_write(0x1FFFF, drv_priv, SGBM_ACMR_BASE);

	ret = write_all_filled(drv_priv, &filled_space);

	list_for_each_entry_safe(tmp, store, &filled_space, list) {
		list_del(&tmp->list);
		kfree(tmp);
	};
	list_for_each_entry_safe(tmp, store, &freespace, list) {
		list_del(&tmp->list);
		kfree(tmp);
	};

	return ret;
};

int hwsgbm_unmap_vram(struct hwsgbm_ctx *ctx)
{
	struct hwsgbm_priv *drv_priv = ctx->drv_priv;
	gen_pool_free(drv_priv->vrampool,
		      (unsigned long)drv_priv->vram_virt,
		      drv_priv->vram_len);
	drv_priv->vram_virt = NULL;
	drv_priv->vram_phys = 0;
	drv_priv->vram_len = 0;
	drv_priv->srcl0_ofst = 0;
	drv_priv->srcl1_ofst = 0;
	drv_priv->srcr0_ofst = 0;
	drv_priv->srcr1_ofst = 0;
	drv_priv->disp0_ofst = 0;
	drv_priv->disp1_ofst = 0;
	return 0;
}
