/*
 * Simple kernel driver that allows for client drivers to control ELVEES VDMA.
 * The driver can be used only with one device.
 *
 * Copyright 2016 ELVEES NeoTek JSC
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

#include <linux/module.h>
#include <linux/of_platform.h>
#include <linux/vdma.h>

#define VDMA_NR_CHANNELS              16
#define VDMA_CHANNEL_OFFSET(n)        ((n) * 0x80)

/* TODO: Specify ELVEES VDMA version */

static struct vdma_pdata {
	struct device *dev;
	void __iomem *regs;
	/* bitmask for requested VDMA channels */
	DECLARE_BITMAP(requested_chans, VDMA_NR_CHANNELS);
} pdata;

int vdma_request_channel(void)
{
	int idx = 0;

	idx = find_first_zero_bit(pdata.requested_chans, VDMA_NR_CHANNELS);
	if (idx == VDMA_NR_CHANNELS) {
		dev_err(pdata.dev, "Can not find free channel\n");
		return -EBUSY;
	}

	set_bit(idx, pdata.requested_chans);

	dev_dbg(pdata.dev, "Channel %d request detected", idx);
	return idx;
}
EXPORT_SYMBOL(vdma_request_channel);

void vdma_release_channel(int channel)
{
	clear_bit(channel, pdata.requested_chans);
	dev_dbg(pdata.dev, "Channel %d released", channel);
}
EXPORT_SYMBOL(vdma_release_channel);

void vdma_write(int channel, u32 reg, u32 data)
{
	iowrite32(data, pdata.regs + VDMA_CHANNEL_OFFSET(channel) + reg);
}
EXPORT_SYMBOL(vdma_write);

u32 vdma_read(int channel, u32 reg)
{
	return ioread32(pdata.regs + VDMA_CHANNEL_OFFSET(channel) + reg);
}
EXPORT_SYMBOL(vdma_read);

static u32 vdma_get_mask(unsigned int channels, int cids[])
{
	int i;
	u32 mask = 0;

	for (i = 0; i < channels; ++i)
		mask |= 1 << cids[i];

	return mask;
}

void vdma_enable(unsigned int channels, int cids[])
{
	const u32 mask = vdma_get_mask(channels, cids);
	u32 val = ioread32(pdata.regs + VDMA_ENABLE);

	iowrite32(val | mask, pdata.regs + VDMA_ENABLE);
}
EXPORT_SYMBOL(vdma_enable);

void vdma_wait_all(unsigned int channels, int cids[])
{
	const u32 mask = vdma_get_mask(channels, cids);
	u32 val;

	do {
		val = ioread32(pdata.regs + VDMA_DONE);
	} while ((val & mask) != mask);
}
EXPORT_SYMBOL(vdma_wait_all);

void vdma_wait_all_alt(unsigned int channels, int cids[])
{
	unsigned int i, done;
	u32 val;

	do {
		for (i = 0, done = 0; i < channels; ++i) {
			val = readl_relaxed(pdata.regs +
					    VDMA_CHANNEL_OFFSET(cids[i]) +
					    VDMA_CHAN_DONE);
			if (val)
				done++;
		}
	} while (done != channels);
}
EXPORT_SYMBOL(vdma_wait_all_alt);

static inline u32 vdma_read_relaxed(int channel, u32 reg)
{
	return readl_relaxed(pdata.regs + VDMA_CHANNEL_OFFSET(channel) + reg);
}

void vdma_dump_channel(char const *const level, int channel)
{
	/* These are first, because it is often very important to read them
	 * early. */
	const u32 aecur = vdma_read_relaxed(channel, VDMA_CHAN_AECUR),
		  aicur = vdma_read_relaxed(channel, VDMA_CHAN_AICUR),
		  csr = vdma_read_relaxed(channel, VDMA_CHAN_CSR);

	const u32 a0e = vdma_read_relaxed(channel, VDMA_CHAN_A0E),
		  a0i = vdma_read_relaxed(channel, VDMA_CHAN_A0I),
		  acnt = vdma_read_relaxed(channel, VDMA_CHAN_ACNT),
		  bccnt = vdma_read_relaxed(channel, VDMA_CHAN_BCCNT),
		  hvecnt = vdma_read_relaxed(channel, VDMA_CHAN_HVECNT),
		  hvicnt = vdma_read_relaxed(channel, VDMA_CHAN_HVICNT);

	const u32 beidx = vdma_read_relaxed(channel, VDMA_CHAN_BEIDX),
		  ceidx = vdma_read_relaxed(channel, VDMA_CHAN_CEIDX),
		  heidx = vdma_read_relaxed(channel, VDMA_CHAN_HEIDX),
		  veidx = vdma_read_relaxed(channel, VDMA_CHAN_VEIDX),
		  biidx = vdma_read_relaxed(channel, VDMA_CHAN_BIIDX),
		  ciidx = vdma_read_relaxed(channel, VDMA_CHAN_CIIDX),
		  hiidx = vdma_read_relaxed(channel, VDMA_CHAN_HIIDX),
		  viidx = vdma_read_relaxed(channel, VDMA_CHAN_VIIDX);

	const u32 cfg = vdma_read_relaxed(channel, VDMA_CHAN_CFG),
		  ainit = vdma_read_relaxed(channel, VDMA_CHAN_AINIT);

	printk("%sVDMA channel #%d: {\n", level, channel);

	printk("%sa0e = %08x, aecur = %08x, a0i = %08x, aicur = %08x,\n",
	       level, a0e, aecur, a0i, aicur);

	printk("%sbeidx = %u, ceidx = %u, heidx = %u, veidx = %u,\n",
	       level, beidx, ceidx, heidx, veidx);

	printk("%sbiidx = %u, ciidx = %u, hiidx = %u, viidx = %u,\n",
	       level, biidx, ciidx, hiidx, viidx);

	printk("%sacnt = %lu, arld = %lu, bcnt = %lu, brld = %lu, "
	       "ccnt = %lu, crld = %lu,\n", level,
	       FIELD_GET(VDMA_ACNT_ACNT, acnt),
	       FIELD_GET(VDMA_ACNT_ARLD, acnt),
	       FIELD_GET(VDMA_BCCNT_BCNT, bccnt),
	       FIELD_GET(VDMA_BCCNT_BRLD, bccnt),
	       FIELD_GET(VDMA_BCCNT_CCNT, bccnt),
	       FIELD_GET(VDMA_BCCNT_CRLD, bccnt));

	printk("%shecnt = %lu, herld = %lu, vecnt = %lu, verld = %lu,\n",
	       level, FIELD_GET(VDMA_HVXCNT_HCNT, hvecnt),
	       FIELD_GET(VDMA_HVXCNT_HRLD, hvecnt),
	       FIELD_GET(VDMA_HVXCNT_VCNT, hvecnt),
	       FIELD_GET(VDMA_HVXCNT_VRLD, hvecnt));

	printk("%shicnt = %lu, hirld = %lu, vicnt = %lu, virld = %lu,\n",
	       level, FIELD_GET(VDMA_HVXCNT_HCNT, hvicnt),
	       FIELD_GET(VDMA_HVXCNT_HRLD, hvicnt),
	       FIELD_GET(VDMA_HVXCNT_VCNT, hvicnt),
	       FIELD_GET(VDMA_HVXCNT_VRLD, hvicnt));

	printk("%scfg = %08x, a_init = %08x, csr = %08x }\n",
	       level, cfg, ainit, csr);
}
EXPORT_SYMBOL(vdma_dump_channel);

void vdma_print_descriptor(char const *const level,
			   struct vdma_descriptor const *const vd)
{
	printk("%sVDMA descriptor at %p: {\n", level, vd);
	printk("%sa0e = %08x, aecur = %08x, a0i = %08x, aicur = %08x,\n",
	       level, vd->a0e, vd->aecur, vd->a0i, vd->aicur);
	printk("%sbeidx = %u, ceidx = %u, heidx = %u, veidx = %u,\n",
	       level, vd->beidx, vd->ceidx, vd->heidx, vd->veidx);
	printk("%sbiidx = %u, ciidx = %u, hiidx = %u, viidx = %u,\n",
	       level, vd->biidx, vd->ciidx, vd->hiidx, vd->viidx);
	printk("%sacnt = %u, arld = %u, bcnt = %u, brld = %u, "
	       "ccnt = %u, crld = %u,\n", level,
	       vd->acnt, vd->arld, vd->bcnt, vd->brld, vd->ccnt, vd->crld);
	printk("%shecnt = %u, herld = %u, vecnt = %u, verld = %u,\n",
	       level, vd->hecnt, vd->herld, vd->vecnt, vd->verld);
	printk("%shicnt = %u, hirld = %u, vicnt = %u, virld = %u,\n",
	       level, vd->hicnt, vd->hirld, vd->vicnt, vd->virld);
	printk("%scfg = %08x, a_init = %08x, csr = %08x }\n",
	       level, vd->cfg, vd->a_init, vd->csr);
}
EXPORT_SYMBOL(vdma_print_descriptor);

static int vdma_probe(struct platform_device *pdev)
{
	struct resource *res;

	pdata.dev = &pdev->dev;

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res) {
		dev_err(&pdev->dev, "failed to get memory resource\n");
		return -ENOENT;
	}

	pdata.regs = devm_ioremap_resource(&pdev->dev, res);
	if (IS_ERR(pdata.regs))
		return PTR_ERR(pdata.regs);

	dev_info(&pdev->dev, "ELVEES VDMA @ 0x%p\n", pdata.regs);

	return 0;
}

#ifdef CONFIG_OF
static const struct of_device_id vdma_dt_ids[] = {
	{ .compatible = "elvees,vdma" },
	{}
};
#endif

static struct platform_driver vdma_driver = {
	.driver = {
		.name = "vdma",
		.of_match_table = of_match_ptr(vdma_dt_ids),
	},
};

/* module_platform_driver_probe() is used to disallow driver unbind */
module_platform_driver_probe(vdma_driver, vdma_probe);

MODULE_DESCRIPTION("ELVEES VDMA driver");
MODULE_LICENSE("GPL");
MODULE_VERSION("0.1");
MODULE_AUTHOR("Dmitriy Zagrebin <dzagrebin@elvees.com>");
