/*
 * ELVEES VDMA test module
 *
 * Copyright 2016 ELVEES NeoTek JSC
 * Copyright 2017 RnD Center "ELVEES", JCS
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#define pr_fmt(fmt) KBUILD_MODNAME ": " fmt

#include <linux/dma-mapping.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/random.h>
#include <linux/sizes.h>
#include <linux/vdma.h>
#include <linux/vmalloc.h>

#define VRAM_BASE    0xe8000000
#define VRAM_SIZE    SZ_2M
#define MAX_CHANNELS 16

static unsigned int rchannels = 1;
module_param(rchannels, uint, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(rchannels, "Number of read channels (default: 1");

static unsigned int wchannels = 1;
module_param(wchannels, uint, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(wchannels, "Number of write channels (default: 1");

/* TODO: Add check that block size is less than 512 KiB (max possible size of
 * transfer usign only ACNT) */
static unsigned int blocksize = SZ_128K;
module_param(blocksize, uint, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(blocksize, "Size of the block (default: 128 KiB)");

static unsigned int iterations = 1;
module_param(iterations, uint, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(iterations, "Number of Iterations (default: 1)");

static bool verify;
module_param(verify, bool, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(verify, "Perform data verification (default: false)");

static bool separate_vram = true;
module_param(separate_vram, bool, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(separate_vram, "Use separate buffers in VRAM (default: true)");

static bool separate_ddr = true;
module_param(separate_ddr, bool, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(separate_ddr, "Use separate buffers in DDR (default: true)");

static unsigned long long per_sec(s64 runtime, unsigned int val)
{
	unsigned long long per_sec = 1000000;

	if (runtime <= 0)
		return 0;

	/* drop precision until runtime is 32-bits */
	while (runtime > UINT_MAX) {
		runtime >>= 1;
		per_sec <<= 1;
	}

	per_sec *= val;
	do_div(per_sec, runtime);
	return per_sec;
}

static unsigned int mib_per_sec(s64 runtime, unsigned int size,
				unsigned int const iterations) {
	return (unsigned int)(per_sec(runtime, (size / SZ_1K) * iterations) /
			      SZ_1K);
}

static void vdma_configure(int cid, enum vdma_dir dir, dma_addr_t external,
			   u32 internal, const unsigned int size,
			   const unsigned int iterations)
{
	u32 val;
	const unsigned int transactions = size >> VDMA_SIZE_16;

	vdma_write(cid, VDMA_CHAN_CSR, 0);

	vdma_write(cid, VDMA_CHAN_A0E, (u32)external);
	vdma_write(cid, VDMA_CHAN_AECUR, (u32)external);
	vdma_write(cid, VDMA_CHAN_A0I, internal);
	vdma_write(cid, VDMA_CHAN_AICUR, internal);

	vdma_write(cid, VDMA_CHAN_ACNT,
		   FIELD_PREP(VDMA_ACNT_ACNT, transactions - 1) |
		   FIELD_PREP(VDMA_ACNT_ARLD, transactions - 1));

	vdma_write(cid, VDMA_CHAN_BCCNT,
		   FIELD_PREP(VDMA_BCCNT_BCNT, iterations % 256) |
		   FIELD_PREP(VDMA_BCCNT_BRLD, 255));
	vdma_write(cid, VDMA_CHAN_HVECNT,
		   FIELD_PREP(VDMA_HVXCNT_HCNT, iterations / 256));
	vdma_write(cid, VDMA_CHAN_HVICNT,
		   FIELD_PREP(VDMA_HVXCNT_HCNT, iterations / 256));

	vdma_write(cid, VDMA_CHAN_BEIDX, -size);
	vdma_write(cid, VDMA_CHAN_BIIDX, -size);
	vdma_write(cid, VDMA_CHAN_HEIDX, -size);
	vdma_write(cid, VDMA_CHAN_HIIDX, -size);

	/* DONE_MSK is required for DMA_DONE register to work correctly.
	 * See rf#5169. */
	val = VDMA_CFG_CYCLE | VDMA_CFG_BRST_AE | VDMA_CFG_BRST_AI |
	      FIELD_PREP(VDMA_CFG_SIZE, VDMA_SIZE_16) | VDMA_CFG_DONE_MSK;

	if (dir)
		val |= VDMA_CFG_DIR;

	vdma_write(cid, VDMA_CHAN_CFG, val);

	vdma_write(cid, VDMA_CHAN_CSR, VDMA_CSR_IMEM_RDY);
}

static unsigned int memcheck(u8 *a, u8 *b, unsigned int size)
{
	u32 errors = 0;

	while (size--) {
		if (*a != *b) {
			errors++;
			pr_debug("Mismatch at %p/%p: %x/%x\n", a, b, *a, *b);
		}
		++a; ++b;
	}

	return errors;
}

static void shuffle(unsigned int a[], const unsigned int len)
{
	int i, j;

	for (i = len - 1; i > 1; --i) {
		/* Not true random in [0; i] but it is not worth it to fix. */
		j = get_random_int() % (i + 1);

		swap(a[i], a[j]);
	}
}

static int vdmatest_buffer_alloc(void **buf, dma_addr_t *dma, unsigned int i)
{
	static void *buf0;
	static dma_addr_t dma0;

	if (separate_ddr || i == 0) {
		*buf = dma_alloc_coherent(NULL, blocksize, dma,
					  GFP_KERNEL | GFP_DMA);
		if (!*buf)
			return -ENOMEM;

		if (i == 0) {
			buf0 = *buf;
			dma0 = *dma;
		}
	} else {
		*buf = buf0;
		*dma = dma0;
	}

	return 0;
}

static void vdmatest_buffer_free(void *buf, dma_addr_t dma, unsigned int i)
{
	if (separate_ddr || i == 0)
		dma_free_coherent(NULL, blocksize, buf, dma);
}

static int __init vdmatest_init(void)
{
	const unsigned int channels = rchannels + wchannels;
	const char *const prefix = (rchannels && wchannels) ? "READ & WRITE" :
				   (rchannels ? "READ" : "WRITE");
	int i, rc = 0, cids[MAX_CHANNELS];
	unsigned int errors[MAX_CHANNELS], totalerrors = 0;
	void *buf[MAX_CHANNELS] = { 0 }, *tmp = 0, *__iomem vram;
	dma_addr_t dmabuf[MAX_CHANNELS];
	unsigned int vrambuf[MAX_CHANNELS] = { 0 };
	ktime_t start;
	s64 time;

	if (channels == 0) {
		pr_err("Zero channels is specified. No test is performed.\n");
		return -EINVAL;
	}

	/* TODO: This check should be moved to parameter callback */
	if (channels > MAX_CHANNELS) {
		pr_err("Too more channels: only %u channels are available\n",
		       MAX_CHANNELS);
		return -EINVAL;
	}

	if (verify && !separate_ddr) {
		pr_warn("Verification is not working with single buffer in DDR\n");
		pr_warn("Disabling verification\n");
		verify = false;
	}

	if (verify && !separate_vram) {
		pr_warn("Verification is not working with single buffer in VRAM\n");
		pr_warn("Disabling verification\n");
		verify = false;
	}

	vram = ioremap(VRAM_BASE, VRAM_SIZE);
	if (!vram) {
		pr_err("Failed to remap VRAM\n");
		return -ENODEV;
	}

	for (i = 0; i < channels; ++i) {
		rc = vdmatest_buffer_alloc(&buf[i], &dmabuf[i], i);
		if (rc) {
			pr_err("Failed to allocate system buffer\n");
			goto err_free_buffers;
		}

		pr_debug("buffer %u: %x@%p physical %pad\n", i, blocksize,
			 buf[i], &dmabuf[i]);
	}

	/* This simplifies channel allocation error handling */
	for (i = 0; i < channels; ++i)
		cids[i] = -ENODEV;

	for (i = 0; i < channels; ++i) {
		cids[i] = vdma_request_channel();
		if (cids[i] < 0) {
			pr_err("Failed to request VDMA channel\n");
			rc = cids[i];
			goto err_free_channels;
		}

		if (separate_vram)
			vrambuf[i] = i;
	}

	/* The shuffle is required to find subtle bugs when one channel copies
	 * data belonging to another buffer. */
	shuffle(vrambuf, channels);

	tmp = vmalloc(blocksize);
	if (!tmp) {
		rc = -ENOMEM;
		goto err_free_channels;
	}

	for (i = 0; i < channels; ++i) {
		if (i <= rchannels)
			get_random_bytes(buf[i], blocksize);
		else {
			get_random_bytes(tmp, blocksize);
			memcpy_toio(vram + vrambuf[i] * blocksize, tmp,
				    blocksize);
		}
	}

	for (i = 0; i < channels; ++i) {
		enum vdma_dir dir = (i <= rchannels) ? VDMA_DIR_TO_VRAM :
				    VDMA_DIR_FROM_VRAM;

		vdma_configure(cids[i], dir, dmabuf[i],
			       VRAM_BASE + vrambuf[i] * blocksize, blocksize,
			       iterations);
	}

	start = ktime_get();
	vdma_enable(channels, cids);
	vdma_wait_all(channels, cids);
	time = ktime_us_delta(ktime_get(), start);

	pr_info("%s (%ur/%uw channels, %u iterations by %u KiB): %llu us, %u MiB/s\n",
		prefix, rchannels, wchannels, iterations, blocksize / SZ_1K,
		time, mib_per_sec(time, blocksize, iterations * channels));

	if (verify) {
		for (i = 0; i < channels; ++i) {
			memcpy_fromio(tmp, vram + vrambuf[i] * blocksize,
				      blocksize);

			errors[i] = memcheck(buf[i], tmp, blocksize);
			totalerrors += errors[i];
		}

		pr_info("TEST %s: %u errors, %ur/%uw channels, %u iterations, %u blocksize\n",
			totalerrors ? "FAILED" : "PASSED", totalerrors,
			rchannels, wchannels, iterations, blocksize);

		if (totalerrors) {
			pr_info("Errors: ");
			for (i = 0; i < channels; ++i)
				pr_cont("%u ", errors[i]);
			pr_cont("\n");

			rc = -EIO;
		}
	}

	vfree(tmp);

err_free_channels:
	for (i = 0; i < channels; ++i)
		if (cids[i] >= 0)
			vdma_release_channel(cids[i]);

err_free_buffers:
	for (i = 0; i < channels; ++i)
		if (buf[i])
			vdmatest_buffer_free(buf[i], dmabuf[i], i);

	iounmap(vram);

	return rc;
}

/* When compiled-in wait for drivers to load first */
late_initcall(vdmatest_init);

static void __exit vdmatest_exit(void)
{
}

module_exit(vdmatest_exit);

MODULE_DESCRIPTION("ELVEES VDMA test module");
MODULE_AUTHOR("Dmitriy Zagrebin <dzagrebin@elvees.com>");
MODULE_AUTHOR("Anton Leontiev <aleontiev@elvees.com>");
MODULE_LICENSE("GPL");
MODULE_VERSION("1.0");
